"""
This module contains a simple test suite parser class that can be used
by other modules to extract information from a test suite
"""
import os, sys, re, json
import logging
import subprocess
from cStringIO import StringIO
from tempfile import NamedTemporaryFile

class InternalParsingError(Exception):
    def __init__(self, filename, lines, failed_line, message):
        self.filename = filename
        self.lines = "".join(["%s | %s" % (line != failed_line and " "*2 or ">"*2, line)
                              for line in lines.splitlines(True)])
        self.message = message

    def __str__(self):
        return ("Error while parsing '%s':\n"
                "%s%s\n") % (self.filename, self.lines, self.message)


class ExternalParsingError(Exception):
    pass


def silent_parser_checker(parser):
    """
    Filtering function that returns false when no element is present
    or an exception was raised
    """
    try:
        data = parser.data
    except InternalParsingError:
        return False

    return bool(data)


def verbose_parser_checker(parser):
    """
    Filtering function that returns false when no element is present
    or an exception was raised
    Additionally, it prints exception information to sys.stderr
    """
    try:
        data = parser.data
    except InternalParsingError as exception:
        print >> sys.stderr, exception
        return False

    return bool(data)


class DataParser(object):
    """
    Data parser to get information about test suites and cases
    """
    _cache = {}

    def __init__(self, filename, valid_keys = None, required_keys = None, element_checking_cb = None):
        self.filename = filename
        self.current_line = None
        self._data = None

        # Description keys may be validated
        self.valid_keys = valid_keys

        # Keys that must be included in the description
        if required_keys:
            self.required_keys = required_keys
        else:
            self.required_keys = []

        # Extra checking might be performed passing a method
        self.element_checking_cb = element_checking_cb


    def __iter__(self):
        """
        Iterate always over data
        """
        return iter(self.data)

    @property
    def data(self):
        """
        Extract information from the suite file
        """
        if not self._data:
            with open(self.filename) as data_file:
                self._data = [self._parse_data(data_lines)
                              for data_lines
                              in self._get_data_lines(data_file)]
        return self._data


    def _parse_data(self, lines):
        """
        Extract information of a single block of lines
        """
        data = {}
        key = value = None
        for line in lines:
            # Set current line to be used to provide
            # parsing error information when needed
            self.current_line = line
            if not line.startswith((' ', '\t')):
                if not ':' in line:
                    raise InternalParsingError(self.filename, "".join(lines), self.current_line,
                                               "Value on multiple lines? Indentation expected\n"
                                               "Key/value pair? Colon separator expected")
                key, value = line.split(':', 1)

                # Remove i18n prefix
                if key[0] == '_':
                    key = key[1:]

                if self.valid_keys and key not in self.valid_keys:
                    raise InternalParsingError(self.filename, "".join(lines), self.current_line,
                                               ("Invalid key: '%s'.\n"
                                                "Expected value: %s or %s\n"
                                                % (key, ", ".join(self.valid_keys[:-1]), self.valid_keys[-1])))
                data[key.strip()] = [value.strip()]
            else:
                if not key:
                    if re.search(r':.+', line):
                        message = "Key/value pair? Line mustn't be indented"
                    else:
                        message = "Value on multiple lines? Previous unindented line should contain a key/value pair"
                    raise InternalParsingError(self.filename, "".join(lines), self.current_line, message)
                data[key].append(line[1:].rstrip())

        for key, values in data.items():
            values = [value for value in values if value]

            # Join all values as separate lines
            # Remove empty line dots
            data[key] = "\n".join([line if line != '.' else ''
                                   for line in values])

        for required_key in self.required_keys:
            if required_key not in data:
                message = "'%s' key not found" % required_key
                raise InternalParsingError(self.filename, "".join(lines), None, message)

        if self.element_checking_cb:
            try:
                self.element_checking_cb(data)
            except ExternalParsingError as exception:
                raise InternalParsingError(self.filename, "".join(lines), None, str(exception))

        return data


    def _get_data_lines(self, lines):
        """
        Return data lines
        """
        data_lines = []
        for line in lines:
            if line.isspace():
                if data_lines:
                    yield data_lines
                    data_lines = []
            else:
                data_lines.append(line)

        if data_lines:
            yield data_lines


class DataEncoder(object):
    """
    Encode metadata objects
    """
    ORDERED_KEYS = ['plugin', 'name', 'depends',
                    'requires', 'command', 'user', 'timeout', 'description']


    def __init__(self, i18n=False):
        self.i18n=i18n


    @staticmethod
    def get(extension, *args, **kwargs):
        encoder_name = ('{ext}DataEncoder'
                        .format(ext=extension.lstrip('.').capitalize()))

        factory = globals().get(encoder_name, TxtDataEncoder)
        return factory(*args, **kwargs)


class TxtDataEncoder(DataEncoder):
    """
    Encode metadata objects to plain text
    """
    def encode(self, data):
        # Force the same ordering as in the user interface
        output = StringIO()

        for key in self.ORDERED_KEYS:
            if key in data:
                value = data[key]

                if self.i18n and key == 'description':
                    key = '_{0}'.format(key)

                lines = [line.rstrip() for line in value.splitlines()]
                if len(lines) > 1:
                    output.write('{0}:\n'.format(key))
                    for line in lines:
                        if line:
                            output.write(' {0}'.format(line))
                        else:
                            output.write(' .')
                        output.write('\n')
                else:
                    if len(value) < 80:
                        output.write('{0}: {1}\n'.format(key, value))
                    else:
                        output.write('{0}:\n {1}\n'.format(key, value))

        output.write('\n')
        output.reset()
        return output.read()


class HtmlDataEncoder(DataEncoder):
    """
    Encode metadata objects to HTML
    """
    def encode(self, data):
        # Force the same ordering as in the user interface
        output = StringIO()
        output.write('<dl>\n')
        for key in self.ORDERED_KEYS:
            if key in data:
                value = data[key]

                lines = [line.rstrip() for line in value.splitlines()]
                if len(lines) > 1:
                    output.write('<dt>{0}</dt>\n'.format(key))
                    output.write('<dd>\n')
                    for line in lines:
                        if line:
                            output.write(' {0}'.format(line))
                        else:
                            output.write(' .')
                        output.write('<br/>\n')
                    output.write('</dd>\n')
                else:
                    output.write('<dt>{0}</dt>\n'
                                 '<dd>{1}</dd>\n'.format(key, value))
        output.write('</dl>\n')
        output.reset()
        return output.read()


class Reporter(object):
    """
    Write a report of selected data
    """
    INDEX_DEPTH = 0

    def __init__(self, destination, tree):
        self.destination = destination
        self.tree = tree


    @staticmethod
    def get(report_type, *args, **kwargs):
        report_type_to_factory = {'Text file': 'TxtReporter',
                                  'XLS file': 'XlsReporter',
                                  'Single HTML file': 'SingleHtmlReporter',
                                  'Multiple HTML files': 'MultipleHtmlReporter'}

        factory_name = report_type_to_factory[report_type]
        factory = globals()[factory_name]
        return factory(*args, **kwargs)


class XlsReporter(Reporter):
    """
    Write a report of selected data
    """
    def write_report(self):
        """
        Write XLS report
        """
        with NamedTemporaryFile(delete=False) as f:
            f.write(json.dumps(self.tree, indent=4))

        reporter = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                'script',
                                'testplanxls.pl')

        try:
            output = subprocess.check_output(
                ["perl", reporter, f.name, self.destination],
                stderr=subprocess.STDOUT)
            logging.info("XLS Test Plan generated successfully %s" % output)
        except subprocess.CalledProcessError, e:
            logging.error("XLS Test Plan:\n%s" % e.output)

        os.unlink(f.name)


class TxtReporter(Reporter):
    """
    Write a report of selected data
    """
    def _get_index(self, parent=None, depth=0):
        """
        Get tree hierarchy for a node
        """
        output = StringIO()

        if parent:
            iterator = iter(parent['children'])
        else:
            iterator = iter(self.tree)

        for node in iterator:
            description = node['description']

            if isinstance(description, dict):
                name = description['name']
            else:
                name = description

            output.write('{indent}- {name}\n'
                         .format(indent=' '*depth*2,
                                 name=name))

            if node['children'] and depth < self.INDEX_DEPTH:
                output.write(self._get_index(node, depth+1))

        output.write('\n')
        return output.getvalue()


    def _get_node(self, node):
        """
        Get node report
        """
        output = StringIO()

        description = node['description']

        if isinstance(description, dict):
            name = description['name']
        else:
            name = description

        output.write('{name}\n'
                     '{underline}\n'
                     .format(name=name,
                             underline='-'*len(name)))

        if node['children']:
            output.write(self._get_index(node))
            for child in node['children']:
                output.write(self._get_node(child))
        else:
            output.write(TxtDataEncoder().encode(description))

        return output.getvalue()


    def write_report(self):
        """
        Write text report
        """
        output = StringIO()
        output.write('Test Plan\n'
                     '---------\n')

        output.write(self._get_index())
        for node in self.tree:
            output.write(self._get_node(node))

        with open(self.destination, 'w') as f:
            f.write(output.getvalue())


class SingleHtmlReporter(Reporter):
    """
    Write a report of selected data
    """
    def _get_index(self, parent=None, depth=0):
        """
        Get tree hierarchy for a node
        """
        output = StringIO()

        if parent:
            iterator = iter(parent['children'])
        else:
            iterator = iter(self.tree)

        output.write('<ul>\n')
        for node in iterator:

            description = node['description']

            if isinstance(description, dict):
                name = description['name']
            else:
                name = description

            output.write('<li>{name}</li>\n'
                         .format(indent=' '*depth*2,
                                 name=name))

            if node['children'] and depth < self.INDEX_DEPTH:
                output.write(self._get_index(node, depth+1))

        output.write('</ul>\n')
        return output.getvalue()


    def _get_node(self, node):
        """
        Get node report
        """
        output = StringIO()

        description = node['description']

        if isinstance(description, dict):
            name = description['name']
        else:
            name = description

        output.write('<h2>{name}</h2>\n'
                     .format(name=name))

        if node['children']:
            output.write(self._get_index(node))
            for child in node['children']:
                output.write(self._get_node(child))
        else:
            output.write(HtmlDataEncoder().encode(description))

        return output.getvalue()


    def write_report(self):
        """
        Write HTML report
        """
        output = StringIO()
        output.write('<h1>Test Plan</h1>\n')

        output.write(self._get_index())
        for node in self.tree:
            output.write(self._get_node(node))

        with open(self.destination, 'w') as f:
            f.write(output.getvalue())


class MultipleHtmlReporter(Reporter):
    """
    Write a report of selected data
    """
    def _get_index(self, parent=None, depth=0):
        """
        Get tree hierarchy for a node
        """
        output = StringIO()

        if parent:
            iterator = iter(parent['children'])
            parent_description = parent['description']
            if isinstance(parent_description, dict):
                parent_name = parent_description['name']
            else:
                parent_name = parent_description
            directory = parent_name.replace(os.path.sep, '-')
        else:
            iterator = iter(self.tree)
            directory = ''

        output.write('<ul>\n')
        for node in iterator:

            description = node['description']

            if isinstance(description, dict):
                name = description['name']
            else:
                name = description

            filename = os.path.join(directory, '{name}.html'
                                    .format(name=name.replace(os.path.sep, '-')))
            output.write('<li><a href="{filename}">{name}</a></li>\n'
                         .format(filename=filename,
                                 name=name))

            if node['children'] and depth < self.INDEX_DEPTH:
                output.write(self._get_index(node, depth+1))

        output.write('</ul>\n')
        return output.getvalue()


    def _write_node(self, parent_directory, node):
        """
        Write node report
        """
        description = node['description']
        if isinstance(description, dict):
            name = description['name']
        else:
            name = description

        filename = os.path.join(parent_directory,
                                '{name}.html'
                                .format(name=name.replace(os.path.sep, '-')))

        with open(filename, 'w') as f:
            f.write('<h1>{name}</h1>\n'
                         .format(name=name))

            if node['children']:
                f.write(self._get_index(node))
            else:
                f.write(HtmlDataEncoder().encode(description))

        if node['children']:
            directory = os.path.join(parent_directory,
                                     name.replace(os.path.sep, '-'))
            os.mkdir(directory)

            for child in node['children']:
                self._write_node(directory, child)


    def write_report(self):
        """
        Write HTML report
        """
        os.mkdir(self.destination)

        filename = os.path.join(self.destination, 'index.html')
        with open(filename, 'w') as f:
            f.write('<h1>Test Plan</h1>\n')
            f.write(self._get_index())

        for node in self.tree:
            self._write_node(self.destination, node)
