"""
checkbox-editor main window module
"""
import pygtk
pygtk.require("2.0")
import gtk, glib

# Use threads
gtk.gdk.threads_init()

import sys, os, re, subprocess, shutil
import logging, logging.handlers
from optparse import OptionParser

import apt

from .about import AboutDialog
from .preferences import Preferences, PreferencesDialog
from .model import TreeStore
from .treeview import Treeview
from .form import Form
from .statusbar import StatusBar
from .util import CommandLauncher, ExceptionHandled, exception_handler, MessageDialogRunner, report_bug
from .validation import JobValidator
from .template import get_available_templates
from . import build


class Editor(object):
    """
    Checkbox editor application that makes it easier to edit test job descriptions
    """
    PACKAGER = '/usr/bin/packager-gtk'

    instance = None

    @classmethod
    def get(cls):
        if not cls.instance:
            cls.instance = Editor()

        return cls.instance


    def __init__(self):
        """
        Create user interface widgets
        """
        build_message = 'Checkbox Editor - {version}'.format(version=build.version)
        self.options = self._parse_command_line()
        if self.options.version:
            print build_message
            sys.exit(0)

        self.validator = JobValidator()

        # Replace default exception handler
        # with one that uses logging and Gtk
        sys.excepthook = exception_handler

        self._configure_logging(self.options.log_level,
                                self.options.log_file)
        logging.info(build_message)
        logging.info('checkbox_editor package located at: {0}'
                     .format(os.path.dirname(__file__)))
        logging.debug('Options parsed: {0}'.format(self.options))

        self.preferences = Preferences()
        self.menu = {}
        self.builder = gtk.Builder()

        ui_file = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                               'glade', 'editor.glade')
        self.builder.add_from_file(ui_file)

        self.form = Form(self.builder, self.preferences)
        self.statusbar = StatusBar(self.builder.get_object('statusbar'))
        self.treeview = Treeview(self.builder, self.form, self.statusbar)

        window = self.builder.get_object('window')

        self._init_text_buffers()
        self._set_accel_group(window)
        model = self._init_treeview(self.options)
        whitelist = model.whitelist
        self._init_job_depends_combobox(model)

        # Enable packager menuitem if binary is found
        packager_menuitem = self.builder.get_object('packager_menuitem')
        packager_menuitem.set_sensitive(os.path.isfile(self.PACKAGER))

        self.builder.connect_signals({
                'window_delete_event_cb': self.window_delete_event_cb,
                'window_destroy_cb': self.window_destroy_cb,
                'quit_menuitem_activate_cb': self.quit_menuitem_activate_cb,
                'new_menuitem_activate_cb': self.new_menuitem_activate_cb,
                'open_menuitem_activate_cb': self.open_menuitem_activate_cb,
                'close_menuitem_activate_cb': self.close_menuitem_activate_cb,
                'revert_menuitem_activate_cb': self.revert_menuitem_activate_cb,
                'save_menuitem_activate_cb': self.save_menuitem_activate_cb,
                'save_as_menuitem_activate_cb': self.save_as_menuitem_activate_cb,
                'open_whitelist_menuitem_activate_cb': self.open_whitelist_menuitem_activate_cb,
                'save_whitelist_menuitem_activate_cb': self.save_whitelist_menuitem_activate_cb,
                'save_whitelist_as_menuitem_activate_cb': self.save_whitelist_as_menuitem_activate_cb,
                'export_whitelist_menuitem_activate_cb': self.export_whitelist_menuitem_activate_cb,
                'preferences_menuitem_activate_cb': self.preferences_menuitem_activate_cb,
                'cut_menuitem_activate_cb': self.treeview.cut_menuitem_activate_cb,
                'copy_menuitem_activate_cb': self.treeview.copy_menuitem_activate_cb,
                'paste_menuitem_activate_cb': self.treeview.paste_menuitem_activate_cb,
                'view_menuitem_toggled_cb': self.treeview.view_menuitem_toggled_cb,
                'whitelist_top_button_clicked_cb': whitelist.top_button_clicked_cb,
                'whitelist_up_button_clicked_cb': whitelist.up_button_clicked_cb,
                'whitelist_down_button_clicked_cb': whitelist.down_button_clicked_cb,
                'whitelist_bottom_button_clicked_cb': whitelist.bottom_button_clicked_cb,
                'execute_menuitem_activate_cb': self.execute_menuitem_activate_cb,
                'push_menuitem_activate_cb': self.push_menuitem_activate_cb,
                'pull_menuitem_activate_cb': self.pull_menuitem_activate_cb,
                'commit_menuitem_activate_cb': self.commit_menuitem_activate_cb,
                'show_problems_menuitem_activate_cb': self.show_problems_menuitem_activate_cb,
                'statistics_menuitem_activate_cb': self.statistics_menuitem_activate_cb,
                'package_requirements_menuitem_activate_cb': self.package_requirements_menuitem_activate_cb,
                'packager_menuitem_activate_cb': self.packager_menuitem_activate_cb,
                'contents_menuitem_activate_cb': self.contents_menuitem_activate_cb,
                'report_problem_menuitem_activate_cb': self.report_problem_menuitem_activate_cb,
                'about_menuitem_activate_cb': self.about_menuitem_activate_cb,
                'job_buffer_insert_text_cb': self.form.job_buffer_insert_text_cb,
                'job_command_changed_cb': self.form.job_command_changed_cb,
                'job_command_edit_clicked_cb': self.form.job_command_edit_clicked_cb,
                'job_command_execute_clicked_cb': self.form.job_command_execute_clicked_cb,
                'value_changed_cb_after': self.form.value_changed_cb_after,
                'focus_out_event_cb': self.form.focus_out_event_cb,
                'add_toolbutton_clicked_cb': self.treeview.add_toolbutton_clicked_cb,
                'remove_toolbutton_clicked_cb': self.treeview.remove_toolbutton_clicked_cb,
                'treeview_button_press_event_cb': self.treeview.treeview_button_press_event_cb,
                'key_press_event_cb': self.treeview.key_press_event_cb,
                })

        window.show_all()


    def _init_treeview(self, options):
        """
        Initialize treeview model
        """
        self.preferences.load(options.directory)

        model = TreeStore(self.builder, self.preferences)
        treeview = self.builder.get_object('treeview')
        treeview.set_model(model)
        if (options.directory
            and model.is_valid_directory(options.directory)):
            self._open_valid_directory(options.directory)

        targets = [('Reorder job', gtk.TARGET_SAME_WIDGET, 0)]
        treeview.enable_model_drag_source(gtk.gdk.BUTTON1_MASK,
                                          targets,
                                          gtk.gdk.ACTION_MOVE)
        treeview.enable_model_drag_dest(targets,
                                        gtk.gdk.ACTION_MOVE)
        treeview.connect('drag-data-received', self.treeview.drag_data_received_cb)

        renderer = gtk.CellRendererToggle()
        renderer.connect('toggled', self.treeview.toggled_cb)
        column = gtk.TreeViewColumn('', renderer,
                                    active=model.SELECTED)
        treeview.append_column(column)

        renderer = gtk.CellRendererText()
        renderer.set_properties(editable=True, editable_set=True)
        renderer.connect('edited', self.treeview.row_edited_cb)
        column = gtk.TreeViewColumn('Name', renderer,
                                    font_desc=model.FONT_DESCRIPTION,
                                    markup=model.MARKUP)
        column.set_sort_indicator(True)
        column.set_sort_order(gtk.SORT_ASCENDING)
        column.set_sort_column_id(0)
        treeview.append_column(column)

        selection = treeview.get_selection()
        selection.set_mode(gtk.SELECTION_MULTIPLE)

        # Callback to update form information
        # when a new item is selected
        selection.connect('changed', self.treeview.selection_changed_cb)

        # Callback to display popup menu
        # if right button was clicked previously
        selection.connect_after('changed', self.treeview.selection_changed_cb_after)

        return model


    def _init_text_buffers(self):
        """
        Initialize text buffers by setting
        a data field with their parent
        """
        widgets = [widget for widget in self.builder.get_objects()
                   if type(widget) == gtk.TextView]

        for widget in widgets:
            text_buffer = widget.get_buffer()
            text_buffer.set_data('parent', widget)


    def _init_job_depends_combobox(self, model):
        """
        Initialize job depends combobox
        """
        combobox = self.builder.get_object('job_depends')
        combobox.set_model(model)
        entry = combobox.get_child()
        entry.connect_after('changed', self.form.value_changed_cb_after)
        entry.set_data('parent', combobox)
        entry.set_name('job_depends_entry')


    def _set_accel_group(self, window):
        """
        Set accelerators for common actions
        """
        group = gtk.AccelGroup()
        window.add_accel_group(group)

        accelerators_data = [
            ('new_menuitem', '<Control>n'),
            ('open_menuitem', '<Control>o'),
            ('close_menuitem', '<Control>w'),
            ('quit_menuitem', '<Control>q'),
            ('save_menuitem', '<Control>s'),
            ('save_as_menuitem', '<Shift><Control>s'),
            ('add_menuitem', '<Control>j'),
            ('remove_menuitem', '<Control>r'),
            ('preferences_menuitem', '<Control>p'),
            ('contents_menuitem', '<Control>F1'),
            ]

        for widget_name, keys in accelerators_data:
            menuitem = self.builder.get_object(widget_name)
            accel_key, accel_mod = gtk.accelerator_parse(keys)
            menuitem.add_accelerator("activate",
                                     group,
                                     accel_key,
                                     accel_mod,
                                     gtk.ACCEL_VISIBLE)


    def _configure_logging(self, log_level, log_filename):
        """
        Configure logging based on the configuration
        passed through the command line
        """
        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)

        # Log to sys.stderr using log level
        # passed through command line
        log_handler = logging.StreamHandler()
        formatter = logging.Formatter('%(levelname)s: %(message)s')
        log_handler.setFormatter(formatter)
        log_handler.setLevel(log_level)
        logger.addHandler(log_handler)

        # Log to file using DEBUG log level
        log_filename_exists = os.path.exists(log_filename)
        log_directory = os.path.dirname(log_filename)
        if not os.path.exists(log_directory):
            os.makedirs(log_directory)

        log_handler = (logging.handlers
                       .RotatingFileHandler(log_filename, mode='a+',
                                            backupCount=3))
        formatter = logging.Formatter('%(asctime)s - %(levelname)s: '
                                      '%(message)s')
        log_handler.setFormatter(formatter)
        log_handler.setLevel(logging.DEBUG)
        logger.addHandler(log_handler)

        if log_filename_exists:
            log_handler.doRollover()


    def _parse_command_line(self):
        """
        Parse command line arguments
        """
        default_log_file = os.path.join(os.path.expanduser('~'),
                                        '.checkbox-editor',
                                        'checkbox-editor.log')

        log_levels = ['debug', 'info', 'warning', 'error', 'critical']
        parser = OptionParser()
        parser.add_option('-d', '--directory',
                          help='Shared checkbox directory that should be opened')
        parser.add_option('-l', '--log-level',
                          dest='log_level',
                          choices=log_levels,
                          default='warning',
                          help=('Log level. One of {0} or {1} '
                                '(%default by default)'
                                .format(', '.join(log_levels[:-1]), log_levels[-1])))
        parser.add_option('-f', '--log-file',
                          dest='log_file',
                          default=default_log_file,
                          help=('Filename where logs are to be written '
                                "('%default' by default)"))
        parser.add_option('-v', '--version',
                          action='store_true',
                          help='Print version number and exit')
        options, args = parser.parse_args()

        if options.directory:
            options.directory = os.path.realpath(options.directory)
        options.log_file = os.path.realpath(options.log_file)

        options.log_level = getattr(logging, options.log_level.upper())

        return options


    def set_title(self, directory=None):
        """
        Set main window title based on the opened directory
        """
        title = 'Checkbox Editor'
        if directory:
            title = '{0} - '.format(directory) + title
            if not os.access(directory, os.F_OK | os.W_OK):
                title = '[read only] ' + title
        window = self.builder.get_object('window')
        window.set_title(title)


    def window_delete_event_cb(self, window, event):
        """
        Delete main window
        """
        # Ask user if he wants to save pending changes
        if not self._continue_without_saving():
            return True

        return False


    def window_destroy_cb(self, window):
        """
        Quit from application
        """
        if self.preferences.version_control.get('push_on_close', False):
            branch_url = self.preferences.version_control.get('push_branch', None)
            model = self.builder.get_object('treeview').get_model()
            if model.directory:
                model.vc.push(branch_url)
        gtk.main_quit()


    def quit_menuitem_activate_cb(self, widget):
        """
        Destroy main window
        """
        # Ask user if he wants to save pending changes
        if self._continue_without_saving():
            self.builder.get_object('window').destroy()
        return False


    def new_menuitem_activate_cb(self, widget):
        """
        Create new checkbox-based testing branch
        """
        if not self._continue_without_saving():
            return False

        dialog = gtk.FileChooserDialog('New testing branch',
                                       buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT,
                                                  gtk.STOCK_OPEN, gtk.RESPONSE_ACCEPT),
                                       action = gtk.FILE_CHOOSER_ACTION_CREATE_FOLDER)

        # Set extra widget to make it possible
        # to select template
        hbox = gtk.HBox()
        label = gtk.Label('Template:')
        hbox.pack_start(label, False)

        combobox = gtk.combo_box_new_text()

        dir_templates = dict((dir_template.name, dir_template)
                             for dir_template in get_available_templates())

        for dir_template_name in dir_templates.keys():
            combobox.append_text(dir_template_name)
        combobox.set_active(0)
        hbox.pack_start(combobox)
        dialog.set_extra_widget(hbox)

        dialog.show_all()
        response = dialog.run()

        if response != gtk.RESPONSE_ACCEPT:
            dialog.destroy()
            return False

        directory = dialog.get_filename()
        selected_template_name = combobox.get_active_text()
        dir_template = dir_templates[selected_template_name]
        dir_template.render(directory)

        dialog.destroy()

        self._open_valid_directory(directory)
        return False

    def open_menuitem_activate_cb(self, widget):
        """
        Open checkbox shared directory containing test suites/cases
        """
        if not self._continue_without_saving():
            return False

        dialog = gtk.FileChooserDialog('Open Shared Directory',
                                       buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT,
                                                  gtk.STOCK_OPEN, gtk.RESPONSE_ACCEPT),
                                       action = gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)


        treeview = self.builder.get_object('treeview')
        model = treeview.get_model()

        if model.directory:
            dialog.set_current_folder(model.directory)

        response = dialog.run()
        directory = dialog.get_filename()
        dialog.destroy()

        if response != gtk.RESPONSE_ACCEPT:
            return False

        if model.is_valid_directory(directory):
            self._open_valid_directory(directory)

        return False


    def _open_valid_directory(self, directory):
        """
        Open valid directory once it has been selected/created
        """
        treeview = self.builder.get_object('treeview')
        model = treeview.get_model()

        self.preferences.load(directory)

        for widget_name in ('pull_menuitem', 'pull_toolbutton',
                            'push_menuitem', 'push_toolbutton',
                            'commit_menuitem', 'commit_toolbutton',
                            ):
            widget = self.builder.get_object(widget_name)
            widget.set_sensitive(False)


        if ExceptionHandled == model.load(directory):
            model.clear()
        else:
            for widget_name in ('close_menuitem', 'close_toolbutton',
                                'open_whitelist_menuitem',
                                'show_problems_menuitem', 'statistics_menuitem',
                                'view_jobs_menuitem', 'view_whitelist_menuitem'):
                widget = self.builder.get_object(widget_name)
                widget.set_sensitive(True)

            # Set notebook tab
            whitelist_menuitem = self.builder.get_object('view_whitelist_menuitem')
            notebook = self.builder.get_object('notebook')
            if whitelist_menuitem.get_active():
                notebook.set_current_page(Treeview.PAGE_WHITELIST)
            else:
                notebook.set_current_page(Treeview.PAGE_EMPTY)

            self.set_title(directory)
            self.statusbar._write_tmp_message('Shared directory opened')

            vc = model.vc
            if hasattr(vc, 'revno'):
                revno_label = self.builder.get_object('revno_label')
                revno_label.set_text('rev: {0}'.format(vc.revno()))


    def close_menuitem_activate_cb(self, widget):
        """
        Close checkbox shared directory containing test suites/cases
        """
        if not self._continue_without_saving():
            return False

        treeview = self.builder.get_object('treeview')
        model = treeview.get_model()
        model.clear()

        notebook = self.builder.get_object('notebook')
        notebook.set_current_page(Treeview.PAGE_EMPTY)

        logging.info("Closing directory: %s" % model.directory)

        self.preferences.load()

        widget_names = ('close_menuitem', 'close_toolbutton',
                        'open_whitelist_menuitem', 'save_whitelist_menuitem',
                        'save_whitelist_as_menuitem', 'export_whitelist_menuitem',
                        'pull_menuitem', 'pull_toolbutton',
                        'push_menuitem', 'push_toolbutton',
                        'commit_menuitem', 'commit_toolbutton',
                        'show_problems_menuitem', 'statistics_menuitem',
                        'view_jobs_menuitem', 'view_whitelist_menuitem')
        for widget_name in widget_names:
            self.builder.get_object(widget_name).set_sensitive(False)

        self.set_title()
        self.statusbar._write_tmp_message('Shared directory closed')
        revno_label = self.builder.get_object('revno_label')
        revno_label.set_text('')

        return False


    def save_menuitem_activate_cb(self, widget):
        """
        Save changes to files
        """
        model = self.builder.get_object('treeview').get_model()

        # Warn user about any potential problem not yet solved
        problems = self.validator.get_problems(model)
        if problems:
            response = self.validator.display_problems(problems)
            if response != gtk.RESPONSE_YES:
                return False

        # Save all files that have been modified
        saved = model.save()
        if saved is True:
            self.statusbar._write_tmp_message('Changes saved')

        return False


    def save_as_menuitem_activate_cb(self, widget):
        """
        Save changes to a different directory
        """
        model = self.builder.get_object('treeview').get_model()

        # Warn user about any potential problem not yet solved
        problems = self.validator.get_problems(model)
        if problems:
            response = self.validator.display_problems(problems)
            if response != gtk.RESPONSE_YES:
                return False

        # Choose a destination directory
        dialog = gtk.FileChooserDialog('Select Shared Directory',
                                       buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT,
                                                  gtk.STOCK_OPEN, gtk.RESPONSE_ACCEPT),
                                       action = gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)

        while True:
            response = dialog.run()
            directory = dialog.get_filename()

            if response != gtk.RESPONSE_ACCEPT:
                dialog.destroy()
                return False

            if not os.path.isdir(directory):
                message = ("Selected directory doesn't seem to be a valid "
                           'directory:\n{0}\n\n'
                           'Please make sure that the path <b>exists</b> '
                           "and that it's a directory"
                           .format(directory))
                MessageDialogRunner('Invalid directory', message).run()

            elif not os.access(directory, os.F_OK | os.R_OK | os.W_OK | os.X_OK):
                message = ('Not enough <b>permissions</b> '
                           'to read/write/execute selected directory:\n{0}\n\n'
                           .format(directory))
                MessageDialogRunner('Insuficcient permissions', message).run()

            elif bool(os.listdir(directory)):
                message = ("Selected directory isn't <b>empty</b>:\n{0}\n\n"
                           'Please select an empty directory, '
                           'so that it can be populated with new data properly'
                           .format(directory))
                MessageDialogRunner('Not empty directory', message).run()
            else:
                break
        dialog.destroy()

        model = self.builder.get_object('treeview').get_model()
        model.save_as(directory)
        self.set_title(directory)
        self.statusbar._write_tmp_message('Changes saved')

        return False


    def _continue_without_saving(self):
        """
        Check if there are pending changes and ask user
        if he wants to continue
        """
        # If save_as_menuitem is sensitive,
        # it means there are pending (not saved) changes
        if self.builder.get_object('save_as_menuitem').get_property('sensitive'):
            message = ('Changes have <b>not</b> been saved.\n\n'
                       'Would you like to continue?')
            response = MessageDialogRunner('Changes not saved', message,
                                           type=gtk.MESSAGE_WARNING,
                                           buttons=gtk.BUTTONS_YES_NO).run()

            if response != gtk.RESPONSE_YES:
                return False

        return True


    def revert_menuitem_activate_cb(self, menuitem):
        """
        Revert changes
        """
        treeview = self.builder.get_object('treeview')
        model = treeview.get_model()
        model.revert()

        selection = treeview.get_selection()
        selection.select_path((0,))

        self.statusbar._write_tmp_message('Changes reverted')
        return False


    def open_whitelist_menuitem_activate_cb(self, menuitem):
        """
        Open whitelist file and select rows accordingly
        """
        dialog = gtk.FileChooserDialog('Open Whitelist',
                                       buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT,
                                                  gtk.STOCK_OPEN, gtk.RESPONSE_ACCEPT))

        for name, pattern in (('Checkbox whitelist (*.wl)', '*.wl'),
                              ('All files', '*')):
            filter = gtk.FileFilter()
            filter.set_name(name)
            filter.add_pattern(pattern)
            dialog.add_filter(filter)

        response = dialog.run()
        filename = dialog.get_filename()
        dialog.destroy()

        if response == gtk.RESPONSE_ACCEPT:
            treeview = self.builder.get_object('treeview')
            model = treeview.get_model()
            if ExceptionHandled != model.whitelist.load(filename):
                self.statusbar._write_tmp_message('Whitelist file opened')

        return False


    def save_whitelist_menuitem_activate_cb(self, menuitem):
        """
        Save whitelist to file
        """
        treeview = self.builder.get_object('treeview')
        model = treeview.get_model()
        if ExceptionHandled != model.whitelist.save():
            self.builder.get_object('save_whitelist_menuitem').set_sensitive(False)
            self.statusbar._write_tmp_message('Whitelist saved to {0}'
                                              .format(repr(model.whitelist.filename)))
        return False


    def save_whitelist_as_menuitem_activate_cb(self, menuitem):
        """
        Save whitelist using file chooser dialog
        """
        treeview = self.builder.get_object('treeview')
        model = treeview.get_model()

        dialog = gtk.FileChooserDialog('Select Whitelist file',
                                       buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT,
                                                  gtk.STOCK_SAVE, gtk.RESPONSE_ACCEPT),
                                       action = gtk.FILE_CHOOSER_ACTION_SAVE)
        if model.whitelist.filename:
            dialog.set_filename(model.whitelist.filename)
        elif model.directory:
            dialog.set_current_folder(model.directory)

        for name, pattern in (('Checkbox whitelist (*.wl)', '*.wl'),
                              ('All files', '*')):
            filter = gtk.FileFilter()
            filter.set_name(name)
            filter.add_pattern(pattern)
            dialog.add_filter(filter)

        response = dialog.run()
        filename = dialog.get_filename()
        dialog.destroy()

        if (response == gtk.RESPONSE_ACCEPT
            and ExceptionHandled != model.whitelist.save(filename)):
            self.builder.get_object('save_whitelist_menuitem').set_sensitive(False)
            self.statusbar._write_tmp_message('Whitelist saved to {0}'
                                              .format(repr(filename)))

        return False


    def export_whitelist_menuitem_activate_cb(self, menuitem):
        """
        Export whitelist for documentation purposes
        """
        treeview = self.builder.get_object('treeview')
        model = treeview.get_model()

        dialog = gtk.FileChooserDialog('Select file',
                                       buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT,
                                                  gtk.STOCK_SAVE, gtk.RESPONSE_ACCEPT),
                                       action = gtk.FILE_CHOOSER_ACTION_SAVE)
        if model.directory:
            dialog.set_current_folder(model.directory)


        filter_patterns = {'Text file': '*.txt',
                           'XLS file': '*.xls',
                           'Single HTML file': '*.html',
                           'Multiple HTML files': '*'}
        for name, pattern in filter_patterns.items():
            filter = gtk.FileFilter()
            filter.set_name(name)
            filter.add_pattern(pattern)
            dialog.add_filter(filter)

        response = dialog.run()
        filename = dialog.get_filename()

        # Add default filename extension
        # if not correctly entered
        filter_name = dialog.get_filter().get_name()
        expected_ext = filter_patterns.get(filter_name, '.txt').lstrip('*')
        if filename:
            root, ext = os.path.splitext(filename)
            if not ext or ext != expected_ext:
                filename = root + ext + expected_ext

        dialog.destroy()

        if (response == gtk.RESPONSE_ACCEPT
            and ExceptionHandled != model.whitelist.export(filter_name, filename)):
                self.statusbar._write_tmp_message('Whitelist exported to {0}'
                                                  .format(repr(filename)))
        return False


    def preferences_menuitem_activate_cb(self, menuitem):
        """
        Edit preferences (environment variables assignments)
        """
        logging.info('Editing preferences')
        dialog = PreferencesDialog(self.preferences)
        updated_preferences = dialog.run()
        if updated_preferences:
            # Update excluded files from version control
            model = self.builder.get_object('treeview').get_model()
            vc = model.vc
            if hasattr(vc, 'exclude'):
                vc.exclude = self.preferences.version_control.get('exclude', [])

            if 'environment' in updated_preferences:
                save_dialog_data = ('Environment variables updated',
                                    ('Environment variables have just been <b>updated</b>.\n\n'
                                     'It is recommended to <b>save</b> your changes\n'
                                     'and <b>reload</b> shared directory\n'
                                     'since this new environment variables may affect job file parsing.\n\n'
                                     'Would you like to do so?'))
                reload_dialog_data = ('Environment variables updated',
                                      ('Environment variables have just been <b>updated</b>.\n\n'
                                       'It is recommended to <b>reload</b> shared directory\n'
                                       'since this new environment variables may affect job file parsing.\n\n'
                                       'Would you like to do so?'))
                self._save_and_reload_shared_directory(save_dialog_data, reload_dialog_data)

            self.statusbar._write_tmp_message('Preferences updated')


        return False


    def _save_and_reload_shared_directory(self, save_dialog_data, reload_dialog_data):
        """
        Save changes and reload shared directory
        when some changes have happened to either
        preferences (environment variables updated)
        or branch (new revisions have been pulled)
        """
        close_button = self.builder.get_object('close_toolbutton')
        shared_directory_opened = close_button.get_property('sensitive')

        if shared_directory_opened:
            save_menuitem = self.builder.get_object('save_menuitem')
            save_as_menuitem = self.builder.get_object('save_as_menuitem')
            pending_changes = (save_menuitem.get_property('sensitive')
                               or save_as_menuitem.get_property('sensitive'))

            def reload_shared_directory():
                treeview = self.builder.get_object('treeview')
                self.preferences.problems = []
                model = treeview.get_model()
                model.revert()

                selection = treeview.get_selection()
                selection.select_path((0,))

            if pending_changes:
                title, message = save_dialog_data
                dialog = MessageDialogRunner(title, message,
                                             gtk.MESSAGE_INFO,
                                             gtk.BUTTONS_YES_NO,
                                             default_response=gtk.RESPONSE_YES)
                response = dialog.run()

                if response == gtk.RESPONSE_YES:
                    if save_menuitem.get_property('sensitive'):
                        self.save_menuitem_activate_cb(save_menuitem)
                    elif save_as_menuitem.get_property('sensitive'):
                        self.save_as_menuitem_activate_cb(save_as_menuitem)
                    else:
                        raise ValueError

                    pending_changes = (save_menuitem.get_property('sensitive')
                                       or save_as_menuitem.get_property('sensitive'))
                    if not pending_changes:
                        reload_shared_directory()
            else:
                title, message = reload_dialog_data
                dialog = MessageDialogRunner(title, message,
                                             gtk.MESSAGE_INFO,
                                             gtk.BUTTONS_YES_NO,
                                             default_response=gtk.RESPONSE_YES)
                response = dialog.run()

                if response == gtk.RESPONSE_YES:
                    reload_shared_directory()


    def execute_menuitem_activate_cb(self, widget):
        """
        Execute checkbox using selected rows as a whitelist
        """
        save_button = self.builder.get_object('save_toolbutton')
        pending_changes = save_button.get_property('sensitive')

        # Warn user that changes that have not been saved
        # aren't going to be available in checkbox
        if pending_changes:
            message = ('There are some changes that have <b>not</b> been <b>saved</b>.\n'
                       'Please note that they are not going to be available\n'
                       'in checkbox.\n\n'
                       'Would you like to continue?')
            response = MessageDialogRunner('Changes not saved', message,
                                           type=gtk.MESSAGE_WARNING,
                                           buttons=gtk.BUTTONS_YES_NO).run()
            if response != gtk.RESPONSE_YES:
                return

        treeview = self.builder.get_object('treeview')
        model = treeview.get_model()

        def selected_patterns():
            """
            Return whitelist patterns for selected row
            """
            whitelist_resources = self.preferences.general['whitelist_resources']
            for row in model.get_selected_rows(include_resources=whitelist_resources):
                description = row[model.DESCRIPTION]
                if description and 'name' in description:
                    if ('plugin' in description
                        and description['plugin'].startswith('remote')):
                        # Remote plugins are expected to generate job names
                        # that start with the same prefix as the 'remote'
                        # job used to generate them
                        yield description['name']+'.*'
                    else:
                        yield description['name']

        command = '{0} -w "{1}"'.format(self.preferences.general['checkbox_binary'],
                                        ' '.join(selected_patterns()))

        # Prepare environment variables
        env = os.environ.copy()
        env.update(self.preferences.environment)

        # Remove CHECKBOX_DATA according to configuration
        if (self.preferences.general['checkbox_data_cleanup']
            and 'CHECKBOX_DATA' in env):
            checkbox_data = env['CHECKBOX_DATA']
            if os.path.isdir(checkbox_data):
                logging.debug('Removing previous checkbox data from {0!r}'
                              .format(checkbox_data))
                shutil.rmtree(checkbox_data)

        # Run command in a separate thread
        CommandLauncher(command, env).start()
        return False


    def push_menuitem_activate_cb(self, widget):
        """
        Push changes to remote branch
        """
        model = self.builder.get_object('treeview').get_model()
        branch_url = self.preferences.version_control.get('push_branch', None)
        pushed = model.vc.push(branch_url)

        if pushed:
            for widget_name in ('push_menuitem', 'push_toolbutton'):
                widget = self.builder.get_object(widget_name)
                widget.set_sensitive(False)
            self.statusbar._write_tmp_message('Changes pushed')

        return False


    def pull_menuitem_activate_cb(self, widget):
        """
        Pull changes from remote branch
        """
        model = self.builder.get_object('treeview').get_model()
        branch_url = self.preferences.version_control.get('pull_branch', None)
        pulled = model.vc.pull(branch_url)

        if pulled:
            pull_result = model.vc.pull_result

            if pull_result.new_revno > pull_result.old_revno:
                save_dialog_data = ('New revisions pulled',
                                    ('<b>{0}</b> revisions have been <b>pulled</b>\n\n'
                                     'It is recommended to <b>save</b> your changes\n'
                                     'and <b>reload</b> shared directory\n'
                                     'to make all updates available.\n\n'
                                     'Would you like to do so?'
                                     .format(pull_result.new_revno - pull_result.old_revno)))
                reload_dialog_data = ('New revisions pulled',
                                      ('<b>{0}</b> revisions have been <b>pulled</b>\n\n'
                                       'It is recommended to and <b>reload</b> shared directory\n'
                                       'to make all updates available.\n\n'
                                       'Would you like to do so?'
                                       .format(pull_result.new_revno - pull_result.old_revno)))
                self._save_and_reload_shared_directory(save_dialog_data, reload_dialog_data)
            self.statusbar._write_tmp_message('Changes pulled')


    def commit_menuitem_activate_cb(self, widget):
        """
        Commit changes to local branch
        """
        model = self.builder.get_object('treeview').get_model()
        committed = model.vc.commit()

        if committed:
            for widget_name in ('commit_menuitem', 'commit_toolbutton'):
                widget = self.builder.get_object(widget_name)
                widget.set_sensitive(False)
            for widget_name in ('push_menuitem', 'push_toolbutton'):
                widget = self.builder.get_object(widget_name)
                widget.set_sensitive(True)

            self.statusbar._write_tmp_message('Changes committed')
            vc = model.vc
            if hasattr(vc, 'revno'):
                revno_label = self.builder.get_object('revno_label')
                revno_label.set_text('rev: {0}'.format(vc.revno()))

        return False


    def show_problems_menuitem_activate_cb(self, widget):
        """
        Show problems (if any)
        """
        model = self.builder.get_object('treeview').get_model()

        problems = self.validator.get_problems(model)
        if problems:
            glib.idle_add(self.validator.display_problems, problems, True)
        else:
            MessageDialogRunner('No problems found',
                                'No problems have been found',
                                gtk.MESSAGE_INFO).run()


    def statistics_menuitem_activate_cb(self, widget):
        """
        Display shared directory statistics
        """
        treeview = self.builder.get_object('treeview')
        model = treeview.get_model()
        total, stats = model.get_statistics()
        message = ['Jobs defined by plugin']
        for key in sorted(stats.keys()):
            message.append('- {0}: <b>{1}</b>'.format(key, stats[key]))
        message.append('\n<big>Total: <b>{0}</b></big>'.format(total))
        message = '\n'.join(message)

        MessageDialogRunner('Statistics',
                            message,
                            gtk.MESSAGE_INFO).run()


    def package_requirements_menuitem_activate_cb(self, widget):
        """
        Verify package requirements for selected jobs
        """
        treeview = self.builder.get_object('treeview')
        model = treeview.get_model()

        if any(model.get_selected_rows()):
            label_text = ('The following packages are <b>required</b> '
                          'to run the selected tests:')
            rows = model.get_selected_rows()
        else:
            label_text = ('The following packages are <b>required</b> '
                          'to run the tests:')
            rows = model.get_all_rows()

        # Check package requirements for every row
        package_names_regex = re.compile(r'''package\.name\s*==\s*(['"])(?P<name>.*?)\1''')
        package_names = set()
        for row in rows:
            description = row[model.DESCRIPTION]
            if description:
                requires = description.get('requires')
                if requires:
                    names = [match.group('name')
                             for match in package_names_regex.finditer(requires)]
                    for name in names:
                        package_names.add(name)

        package_names = sorted(package_names)

        cache = apt.cache.Cache()
        not_installed_package_names = \
            [package_name for package_name in package_names
             if package_name not in cache]

        # Display required packages in a dialog
        dialog = gtk.Dialog(title='Package requirements',
                            buttons=(gtk.STOCK_OK, gtk.RESPONSE_ACCEPT))
        dialog.set_border_width(10)

        content_area = dialog.get_content_area()
        content_area.set_spacing(10)

        label = gtk.Label()
        label.set_alignment(0.0, 0.5)
        label.set_markup(label_text)
        content_area.pack_start(label, False)

        scrolled_window = gtk.ScrolledWindow()
        content_area.pack_start(scrolled_window)

        text_buffer = gtk.TextBuffer()
        text_buffer.set_text(', '.join(package_names))

        text_view = gtk.TextView(text_buffer)
        text_view.set_wrap_mode(gtk.WRAP_WORD)
        scrolled_window.add(text_view)

        if not_installed_package_names:
            label = gtk.Label()
            label.set_alignment(0.0, 0.5)
            label.set_markup('The following packages '
                             'are <b>not installed</b>:')
            content_area.pack_start(label, False)

            scrolled_window = gtk.ScrolledWindow()
            content_area.pack_start(scrolled_window)

            text_buffer = gtk.TextBuffer()
            text_buffer.set_text(', '.join(not_installed_package_names))

            text_view = gtk.TextView(text_buffer)
            text_view.set_wrap_mode(gtk.WRAP_WORD)
            scrolled_window.add(text_view)
        else:
            label = gtk.Label()
            label.set_alignment(0.0, 0.5)
            label.set_markup('All the packages are correctly installed')
            content_area.pack_start(label)

        dialog.show_all()
        dialog.run()
        dialog.destroy()


    def packager_menuitem_activate_cb(self, widget):
        """
        Launch packager-gtk and using same directory
        as the one opened as argument
        """
        command = self.PACKAGER

        treeview = self.builder.get_object('treeview')
        model = treeview.get_model()
        directory = model.directory
        if directory:
            command += ' -d {0}'.format(directory)

        CommandLauncher(command).start()


    def contents_menuitem_activate_cb(self, menuitem):
        """
        Display help contents
        """
        help_filename = os.path.join(os.path.dirname(__file__),
                                     "help",
                                     "checkbox-editor.xml")
        subprocess.Popen(["yelp", help_filename])
        return False


    def report_problem_menuitem_activate_cb(self, menuitem):
        """
        Report problem through apport
        """
        report_bug()


    def about_menuitem_activate_cb(self, menuitem):
        """
        Open about dialog
        """
        dialog = AboutDialog()
        dialog.run()
        return False


    def run(self):
        """
        Run gtk mainloop
        """
        gtk.main()
