"""
form module
"""
import pygtk
pygtk.require("2.0")
import gtk, glib

import os, re, logging, shlex

from .model import TreeStore, TreeRowReference
from .util import CommandLauncher, SignalBlocker, MessageDialogRunner
from .validation import JobValidator


class Form(object):
    """
    Takes care of filling job description data correctly
    """
    WIDGET_NAMES = ('plugin', 'name', 'depends', 'requires',
                    'timeout', 'command', 'user', 'description')

    RED_COLOR = gtk.gdk.color_parse('red')
    WHITE_COLOR = gtk.gdk.color_parse('white')


    def __init__(self, builder, preferences):
        self.builder = builder
        self.preferences = preferences
        self.validator = JobValidator()

        for name in self.WIDGET_NAMES:
            widget_name = 'job_{0}'.format(name)
            widget = self.builder.get_object(widget_name)
            widget.set_name(widget_name)


    def fill(self, row):
        """
        Fill form with job data properly
        """
        # Fill form with job data from text file
        widget_names_set = []
        description = row[TreeStore.DESCRIPTION]
        for key, value in description.iteritems():
            widget = self._get_widget(key)

            # Skip unknown keys from description
            if widget is None:
                logging.warning('Unknown key/value in description {0}:{1}'
                                .format(key, value))
                continue

            self._set_widget_value(widget, value)
            widget_names_set.append(key)

        # Reset unset widgets to default values
        widget_names_unset = [widget_name for widget_name in self.WIDGET_NAMES
                              if widget_name not in widget_names_set]

        for widget_name in widget_names_unset:
            widget = self._get_widget(widget_name)
            self._set_widget_default_value(widget)

        # Set widgets as sensitive only if the job file
        # is under the opened directory (LP: #730852)
        filename_row = row
        filename = filename_row[TreeStore.FILENAME]
        while filename is None:
            filename_row = filename_row.parent
            filename = filename_row[TreeStore.FILENAME]
        directory = os.path.dirname(filename)
        opened_directory = self.preferences.directory
        is_under_opened_directory = (
            directory == opened_directory
            or directory.startswith(opened_directory+os.path.sep))

        editable = len(filename) == 0 or is_under_opened_directory
        for widget_name in self.WIDGET_NAMES:
            widget = self._get_widget(widget_name)
            if isinstance(widget, gtk.TextBuffer):
                widget = widget.get_data('parent')
            widget.set_sensitive(editable)

        # Change widget colors
        self._check_invalid_fields(row)


    def _get_widget(self, key):
        """
        Get widget for a given key
        """
        widget_name = 'job_{0}'.format(key)
        widget = self.builder.get_object(widget_name)

        # TextView values are set through their buffer
        if type(widget) == gtk.TextView:
            widget = widget.get_buffer()

        return widget


    def _get_widget_value(self, widget):
        """
        Get new value from widget
        """
        # Get value from widget
        if type(widget) == gtk.Entry:
            value = widget.get_text()
        elif type(widget) == gtk.TextBuffer:
            value = widget.get_text(widget.get_start_iter(),
                                    widget.get_end_iter())
        elif type(widget) == gtk.ComboBox:
            value = widget.get_active_text()
        elif type(widget) == gtk.SpinButton:
            value = widget.get_value_as_int()

            # Make sure that 0 is never present as a value
            value = str(value) if value else None
        else:
            raise Exception, 'Unexpected widget type: {0}'.format(type(widget))

        return value


    def _set_widget_value(self, widget, value):
        """
        Set widget value properly
        """
        # Make sure that entry child widget
        # is used to set value in combobox widgets
        if (type(widget) == gtk.ComboBox
            and widget.get_property('has-entry')):
            widget = widget.get_child()

        # Block 'changed' signal to avoid extra processing
        with SignalBlocker(widget, self.value_changed_cb_after):
            if type(widget) in (gtk.Entry, gtk.TextBuffer):
                widget.set_text(value)
            elif type(widget) == gtk.ComboBox:
                values = [combobox_row[0]
                          for combobox_row in widget.get_model()]

                if not value in values:
                    message = ('Plugin <b>{plugin}</b> not found.'
                               .format(plugin=value))
                    title = 'Plugin not found'

                    dialog = gtk.MessageDialog(type=gtk.MESSAGE_WARNING,
                                               buttons=gtk.BUTTONS_OK)
                    dialog.set_title(title)
                    dialog.set_markup(message)
                    dialog.connect('response', lambda w, d: dialog.destroy())
                    glib.idle_add(dialog.show_all)
                else:
                    widget.set_active(values.index(value))
            elif type(widget) == gtk.SpinButton:
                widget.set_value(float(value))
            else:
                raise Exception, 'Unexpected widget type: {0}'.format(type(widget))


    def _set_widget_default_value(self, widget):
        """
        Set widget default value properly
        """
        # Make sure that entry child widget
        # is used to set value in combobox widgets
        if (type(widget) == gtk.ComboBox
            and widget.get_property('has-entry')):
            widget = widget.get_child()

        # Block 'changed' signal to avoid extra processing
        with SignalBlocker(widget, self.value_changed_cb_after):
            if type(widget) in (gtk.Entry, gtk.TextBuffer):
                widget.set_text('')
                widget.emit('changed')
            elif type(widget) == gtk.ComboBox:
                widget.set_active(0)
                widget.emit('changed')
            elif type(widget) == gtk.SpinButton:
                widget.set_value(0)
                widget.emit('value-changed')
            else:
                raise Exception, 'Unexpected widget type: {0}'.format(type(widget))


    def value_changed_cb_after(self, widget):
        """
        Make sure that updates to values are correctly tracked
        """
        treeview = self.builder.get_object('treeview')
        selection = treeview.get_selection()
        model, paths = selection.get_selected_rows()
        path = paths[0]
        row = model[path]
        iterator = row.iter

        def get_name(widget):
            """
            Get the name from the widget that has been changed
            except for gtk.TextBuffer. In such a case, get name
            from gtk.TextView parent
            """
            if not getattr(widget, 'name', False):
                widget = widget.get_data('parent')
            return widget.name

        name = get_name(widget)
        key = name.split('_', 1)[1]
        description = row[TreeStore.DESCRIPTION]

        value = self._get_widget_value(widget)

        if key == 'plugin':
            # Make sure that the change from local plugin (has children)
            # to any of the other ones is intentional (don't have children)
            if description[key] == 'local' and model.iter_has_child(iterator):
                change_cancelled = self._confirm_plugin_change_from_local(row, key, value)
                if change_cancelled:
                    return False

            # Make sure that job command is set to a sensible value
            # when changing to local plugin
            if value == 'local':
                self.builder.get_object('job_command').set_sensitive(False)
                self.builder.get_object('job_command_edit').set_sensitive(True)

                if ('command' in description
                    and not self.preferences.parse_job_command(description['command'])):
                    for widget_name in ('add_toolbutton', 'add_menuitem'):
                        self.builder.get_object(widget_name).set_sensitive(False)

            else:
                self.builder.get_object('job_command').set_sensitive(True)
                self.builder.get_object('job_command_edit').set_sensitive(False)

                for widget_name in ('add_toolbutton', 'add_menuitem'):
                    self.builder.get_object(widget_name).set_sensitive(True)


        # Update sibling jobs that depend on the job being modified
        elif key == 'name':
            # Update name cache used to detect duplicate names
            old_name = row[TreeStore.DISPLAY_NAME]
            row_reference = TreeRowReference(row.model, row.path)
            if (old_name
                and old_name in model.name_cache
                and row_reference in model.name_cache[old_name]):

                model.name_cache[old_name].remove(row_reference)

                if len(model.name_cache[old_name]) == 1:
                    same_name_reference = model.name_cache[old_name][0]
                    same_name_row = model[same_name_reference.get_path()]
                    if self.validator.is_valid(same_name_row[TreeStore.DESCRIPTION],
                                               getattr(self.preferences, 'extra_validation')):
                        model._set_problem_found(same_name_row.iter, False)
            if value:
                row_reference = TreeRowReference(row.model, row.path)
                model.name_cache[value].append(row_reference)

            # Update dependencies in sibling jobs
            if 'name' in description:
                for sibling_row in row.parent.iterchildren():
                    sibling_description = sibling_row[TreeStore.DESCRIPTION]

                    if (sibling_description is not description
                        and ('depends' in sibling_description
                             and sibling_description['depends'] == description['name'])):

                        if value:
                            sibling_description['depends'] = value
                        else:
                            del sibling_description['depends']

                        # Set column as modified and text style to bold and italic
                        model._set_modified(sibling_row.iter)

        # Update value in job description
        if key in description and not value:
            del description[key]

            # Update name in model to update treeview
            if key == 'name':
                row[TreeStore.DISPLAY_NAME]= ''
        else:
            description[key] = value

            # Update name in model to update treeview
            if key == 'name':
                row[TreeStore.DISPLAY_NAME]=value

                match = re.match('\<span foreground="(?P<color>\w+)"\>.+\</span\>',
                                 row[TreeStore.MARKUP])
                if match:
                    row[TreeStore.MARKUP] = ('<span foreground="{0}">{1}</span>'
                                             .format(match.group('color'), value))
                else:
                    row[TreeStore.MARKUP]=value

        logging.debug('Value changed: (widget: {0!r}, value: {1!r})'
                      .format(name, value))

        # Set red color for invalid fields
        self._check_invalid_fields(row)

        # Set row as modified
        model._set_modified(iterator)

        return False


    def focus_out_event_cb(self, widget, event):
        """
        focus-out-event is used to track job command changes
        with regard to version control
        """
        command = shlex.split(widget.get_text())
        if not command:
            return False

        script_name = command[0]
        script_name = os.path.relpath(self.preferences.expand(script_name),
                                      self.preferences.directory)

        # Take into account that script path may not be complete
        # if scripts directory is included in PATH environment variable
        if not os.path.isfile(script_name):
            new_script_name = os.path.join('scripts', script_name)
            if os.path.isfile(new_script_name):
                script_name = new_script_name

        treeview = self.builder.get_object('treeview')
        model = treeview.get_model()
        if (script_name.startswith('scripts'+os.sep)
            and not script_name in model.vc):
            model.vc.add(script_name)

        return False


    def _check_invalid_fields(self, row):
        """
        Check that all fields contain valid data
        """
        # First problem found will be displayed in the status bar
        statusbar = self.builder.get_object('statusbar')
        context_id = statusbar.get_context_id('problem')
        statusbar.pop(context_id)
        problem_reason = None

        model = row.model
        description = row[TreeStore.DESCRIPTION]
        invalid_fields_data = self.validator.get_invalid_fields(description,
                                                                getattr(self.preferences, 'extra_validation'))
        invalid_field_names = [data[0] for data in invalid_fields_data]

        # Make sure that the same name isn't being used somewhere else
        name = description.get('name')
        if name:
            references = model.name_cache.get(name, [])

            if len(references) > 1:
                invalid_field_names.append('name')
                message = ('{0!r} is already being used as name by multiple jobs'
                           .format(name))
                logging.error(message)
                problem_reason = message

                for reference in references:
                    same_name_row = model[reference.get_path()]
                    model._set_problem_found(same_name_row.iter)

        # Make sure that dependency is set pointing to a job that exists
        dependencies = description.get('depends')
        if dependencies:
            for dependency in dependencies.split():
                if dependency not in model.name_cache:
                    invalid_field_names.append('depends')

                    message =('{0!r} dependency not found'
                              .format(dependency))
                    logging.warning(message)

                    if not problem_reason:
                        problem_reason = message
                    break

        # Add problem reason to statusbar
        if (invalid_fields_data
            and (not problem_reason or 'name' in invalid_field_names)):
            problem_reason = invalid_fields_data[0][1]

        if problem_reason:
            statusbar.push(context_id, problem_reason)

        # Take care of coloring in form view
        valid_field_names = set(self.WIDGET_NAMES).difference(invalid_field_names)
        for field in valid_field_names:
            widget = self._get_widget(field)
            self.set_colors_default(widget)
        for field in invalid_field_names:
            widget = self._get_widget(field)
            self.set_colors_error(widget)

        # Take care of coloring in tree view
        if invalid_field_names:
            model._set_problem_found(row.iter)
        else:
            model._set_problem_found(row.iter, False)


    def _confirm_plugin_change_from_local(self, row, key, value):
        """
        Make sure that the change from local plugin (has children)
        o any of the other ones is intentional (don't have children)
        """
        description = row[TreeStore.DESCRIPTION]

        # Prompt user to confirm change
        message=('Jobs that use the {0!r} plugin cannot contain '
                 'children jobs definitions, so they will <b>removed</b>. '
                 'Is that what you would like to do?'.format(value))
        response = MessageDialogRunner('Cannot have children', message,
                                       type=gtk.MESSAGE_WARNING,
                                       buttons=gtk.BUTTONS_YES_NO).run()
        change_cancelled = False if response == gtk.RESPONSE_YES else True

        if change_cancelled:
            # Restore original value in combobox
            value = description[key]
            combobox = self.builder.get_object('job_plugin')
            combobox_model = combobox.get_model()

            for index, row in enumerate(combobox_model):
                if row[0] == value:
                    # Block 'changed' signal to avoid extra processing
                    with SignalBlocker(combobox, self.value_changed_cb_after):
                        combobox.set_active(index)
                    break
            else:
                raise Exception('Invalid job plugin value')
        else:
            # Gather children row references
            model = row.model
            children_references = [TreeRowReference(model, child_row.path)
                                   for child_row in row.iterchildren()]

            # Remove paths from name_cache
            for child_row in row.iterchildren():
                description = child_row[model.DESCRIPTION]
                child_row_reference = TreeRowReference(child_row.model, child_row.path)
                if description:
                    name = description.get('name')
                    if (name
                        and name in model.name_cache
                        and child_row_reference in model.name_cache[name]):
                        model.name_cache[name].remove(child_row_reference)

            # Remove rows
            for child_reference in children_references:
                child_path = child_reference.get_path()
                children_iterator = model.get_iter(child_path)
                logging.debug('Removing job: {0}'
                              .format(model[children_iterator][TreeStore.DISPLAY_NAME]))
                model.remove(children_iterator)

            # Remember files to delete
            if 'command' in description:
                job_filenames = self.preferences.parse_job_command(description['command'])
                if job_filenames:
                    logging.debug('Files will be deleted: {0}'
                                  .format(job_filenames))
                    model.files_to_remove.extend(job_filenames)

            # Remove filename
            row[TreeStore.FILENAME] = None


        return change_cancelled


    def job_buffer_insert_text_cb(self, text_buffer, location, text, length):
        """
        Capture tab insertion and replace it by spaces
        """
        if text=='\t' and self.preferences.general['replace_tabs_with_spaces']:
            # Stop emission of the current signal
            text_buffer.stop_emission('insert-text')
            # and emit a new one with the right text
            space_string = ' '*self.preferences.general['number_of_spaces']
            text_buffer.emit('insert-text', location,
                             space_string, len(space_string))

        return True


    def job_command_changed_cb(self, entry):
        """
        Update execute button sensitivity
        """
        command = entry.get_text()
        self.builder.get_object('job_command_execute').set_sensitive(bool(command))


    def job_command_edit_clicked_cb(self, button):
        """
        Select a test case file and fill the the suite command data
        """
        treeview = self.builder.get_object('treeview')
        selection = treeview.get_selection()
        model, paths = selection.get_selected_rows()
        path = paths[0]
        row = model[path]
        description = row[model.DESCRIPTION]

        command_entry = self.builder.get_object('job_command')
        command = command_entry.get_text()

        suggested_filename = None
        if command:
            filenames = self.preferences.parse_job_command(command)
            if len(filenames) == 1:
                suggested_filename = filenames[0]

        if suggested_filename is None:
            if 'name' in description:
                suggested_filename = ('{0}.txt.in'.format(description['name'])
                                      .replace(' ', '_'))

                suggested_filename = os.path.join(model.jobs_dir, suggested_filename)

                # Don't suggest a filename that already exists
                if os.path.isfile(suggested_filename):
                    suggested_filename = None

        filename = model.select_valid_job_file(suggested_filename)
        if filename:
            old_filename = row[TreeStore.FILENAME]
            if (old_filename
                and old_filename in model.vc):
                model.append_rename_operation(old_filename, filename)

            # Set filename to be used for children jobs
            row[model.FILENAME] = filename
            for widget_name in ('add_toolbutton', 'add_menuitem'):
                self.builder.get_object(widget_name).set_sensitive(True)

            # Set widgets sensitivity based on selected file
            directory = os.path.dirname(filename)
            opened_directory = self.preferences.directory
            editable = (directory == opened_directory
                        or directory.startswith(opened_directory+os.path.sep))
            for widget_name in self.WIDGET_NAMES:
                widget = self._get_widget(widget_name)
                if isinstance(widget, gtk.TextBuffer):
                    widget = widget.get_data('parent')
                widget.set_sensitive(editable)

            # Substitute environment variables
            filename = self.preferences.substitute(filename)

            # Transform filename in glob pattern
            pattern = re.sub('\.txt(\.in)?', '.txt*', filename)

            new_command = 'cat {0}'.format(pattern)
            command_entry.set_text(new_command)

        return False


    def job_command_execute_clicked_cb(self, button):
        """
        Excute job command to check if it's correct
        """
        env = os.environ.copy()

        # Update environment variables from preferences
        env.update(self.preferences.environment)

        # Add PATH variables according to preferences
        scripts_directories = filter(os.path.isdir,
                                     [os.path.join(directory, 'scripts')
                                      for directory in self.preferences.environment.values()])
        env['PATH'] += ':{0}'.format(':'.join(scripts_directories))

        command_str = self.builder.get_object('job_command').get_text()

        # Run command in a separate thread
        CommandLauncher(command_str, env).start()

        return False


    def set_colors_error(self, widget):
        """
        Set a widget colors to let the user know there's some kind of problem
        This happens when it has a bad or unexpected value
        """
        # Get entry widget from combobox
        if (type(widget) == gtk.ComboBox
            and widget.get_property('has-entry')):
            widget = widget.get_child()

        if type(widget) not in (gtk.Entry, gtk.TextBuffer):
            return

        if type(widget) == gtk.TextBuffer:
            widget = widget.get_data('parent')

        state = (gtk.STATE_NORMAL
                 if widget.get_property('sensitive') else gtk.STATE_INSENSITIVE)

        current_base_color = widget.style.base[state]
        current_text_color = widget.style.text[state]
        if (current_base_color != self.RED_COLOR
            or current_text_color != self.WHITE_COLOR):

            # Store default colors if needed
            for color_type in ('base', 'text'):
                if not widget.get_data('default_{0}'.format(color_type)):
                    default_colors = dict([(key, getattr(widget.style, color_type)[key])
                                           for key in (gtk.STATE_NORMAL, gtk.STATE_INSENSITIVE)])
                    widget.set_data('default_{0}'.format(color_type), default_colors)

            widget.modify_base(state, self.RED_COLOR)
            widget.modify_text(state, self.WHITE_COLOR)


    def set_colors_default(self, widget):
        """
        Set a widget colors to default values
        """
        # Get entry widget from combobox
        if (type(widget) == gtk.ComboBox
            and widget.get_property('has-entry')):
            widget = widget.get_child()

        if type(widget) not in (gtk.Entry, gtk.TextBuffer):
            return

        if type(widget) == gtk.TextBuffer:
            widget = widget.get_data('parent')

        for color_type in ('base', 'text'):
            default_colors = widget.get_data('default_{0}'.format(color_type))
            if default_colors:
                # Check colors for all available widget states
                # not just the current one
                for state, default_color in default_colors.items():
                    default_color = default_colors[state]
                    current_color = getattr(widget.style, color_type)[state]

                    if default_color != current_color:
                        modify_method = getattr(widget, 'modify_{0}'.format(color_type))
                        modify_method(state, default_color)
