"""
job model implementation
"""
import pygtk
pygtk.require("2.0")
import gtk, pango, glib

import os, re, logging, itertools, collections
from copy import copy as shallow_copy
from bzrlib.errors import BzrError

from .data import DataParser, DataEncoder, InternalParsingError
from .vc import VersionControl
from .util import handle_exceptions, ExceptionDialog, ExceptionHandled, MessageDialogRunner
from .validation import JobValidator
from .whitelist import Whitelist


class RenameOverwriteError(Exception):
    def __init__(self, operation_1, operation_2):
        super(RenameOverwriteError, self).__init__()
        self.operation_1 = operation_1
        self.operation_2 = operation_2


    def __str__(self):
        return ('conflict - {0} => {1}, {2} => {3}'
               .format(self.operation_1[0],
                       self.operation_1[1],
                       self.operation_2[0],
                       self.operation_2[1]))


class CancelException:
    pass


class TreeStore(gtk.TreeStore):
    """
    Tree store implementation with some additional functionality to
    load/save data
    """
    # TreeModel column types
    COLUMNS = (str, object, str, bool, bool, object,
               pango.FontDescription, str, bool)

    # TreeModel column indexes
    DISPLAY_NAME = 0
    DESCRIPTION = 1
    FILENAME = 2
    MODIFIED = 3
    CHILDREN_MODIFIED = 4
    CHILDREN_REMOVED = 5
    FONT_DESCRIPTION = 6
    MARKUP = 7
    SELECTED = 8

    DEFAULT_PLUGIN = 'manual'

    def __init__(self, builder, preferences):
        gtk.TreeStore.__init__(self, *self.COLUMNS)

        self.name_cache = collections.defaultdict(list)
        self.builder = builder
        self.preferences = preferences
        self.validator = JobValidator()
        self.problems = []
        self.whitelist = Whitelist(self)

        self.clear()


    def clear(self):
        """
        Initialize model state
        """
        self.directory = None
        self.jobs_dir = None
        self.scripts_dir = None
        self.vc = None
        self.whitelist.clear()

        self.files_to_remove = []
        self.files_to_rename = {}
        gtk.TreeStore.clear(self)

        # Disable all save/revert options
        for widget_name in ['save_menuitem', 'save_toolbutton',
                            'save_as_menuitem', 'revert_menuitem']:
            widget = self.builder.get_object(widget_name)
            widget.set_sensitive(False)


    def get_all_rows(self, parent = None):
        """
        Return all rows
        """
        for row in (parent.iterchildren() if parent else self):
            yield row

            if self.iter_has_child(row.iter):
                for subjob_row in self.get_all_rows(row):
                    yield subjob_row


    def get_selected_rows(self, parent=None, include_resources=False):
        """
        Return all selected rows
        possibly including resources depending on configuration
        """
        if include_resources:
            def has_child_resources(row):
                """
                Return True if any child is a resource
                """
                for child_row in self.get_all_rows(row):
                    description = child_row[self.DESCRIPTION]
                    if (description
                        and description.get('plugin') == 'resource'):
                        return True
                return False

            def is_row_selected(row):
                """
                Filter selected rows
                and resources rows according to configuration
                """
                # Filter selected rows
                if row[self.SELECTED]:
                    return True

                # Filter resources rows
                if self.preferences.general['whitelist_resources']:
                    description = row[self.DESCRIPTION]
                    if description:
                        plugin = description.get('plugin')
                        if plugin == 'resource':
                            return True
                        else:
                            return has_child_resources(row)

                return False
        else:
            def is_row_selected(row):
                if row[self.SELECTED]:
                    return True

        return self.get_matching_rows(is_row_selected, parent)


    def get_matching_rows(self, matching_func, parent=None):
        """
        Return all rows that match a selection criteria
        """
        for row in self.get_all_rows(parent):
            if matching_func(row):
                yield row


    def get_selected_description_tree(self, parent=None):
        """
        Return all selected row description in a tree hierarchy
        """
        nodes = []
        for row in (parent.iterchildren() if parent else self):
            if not row[self.SELECTED]:
                continue

            node = {'description': row[self.DESCRIPTION] or row[self.DISPLAY_NAME],
                    'children': []}

            if self.iter_has_child(row.iter):
                node['children'] = self.get_selected_description_tree(row)
            nodes.append(node)

        return nodes


    def get_statistics(self, parent=None):
        """
        Get statistics for a given tree/subtree
        """
        stats = collections.defaultdict(int)
        total = 0
        for row in self.get_all_rows(parent):
            description = row[self.DESCRIPTION]
            if description:
                total += 1
                plugin = description.get('plugin')
                stats[plugin] += 1
        return total, stats


    def revert(self):
        """
        Discard changes and reload data from directory
        """
        logging.info('Reverting changes')
        self.load(self.directory)


    def is_valid_directory(self, directory):
        """
        Return True if it's a valid directory and data should be loaded
        In case of not enough permissions, the user might be prompted
        to use read-only mode
        """
        # Make sure directory exists
        if not os.path.isdir(directory):
            message = ("Selected directory doesn't seem to be a valid "
                       'shared checkbox directory:\n{0}\n\n'
                       'Please make sure that the path <b>exists</b> '
                       "and that it's a directory"
                       .format(directory))
            MessageDialogRunner('Invalid directory', message).run()
            return False

        # Make sure that it's possible to access to selected directory
        if not os.access(directory, os.F_OK | os.R_OK | os.X_OK):
            message = ('Not enough <b>permissions</b> to access selected directory:\n{0}\n\n'
                       'Please change permissions accordingly or run checkbox-editor '
                       'with from a different user account with sufficient permissions.'
                       .format(directory))
            MessageDialogRunner('Insuficcient permissions', message).run()
            return False

        # Make sure that directory contains a jobs subdirectory
        jobs_dir = os.path.join(directory, 'jobs')
        if not os.path.isdir(jobs_dir):
            message = ("Selected directory doesn't seem to be a valid "
                       'shared checkbox directory:\n{0}\n\n'
                       'Please make sure to select a directory '
                       'that at least contains a <b>jobs</b> subdirectory.'
                       .format(directory))
            MessageDialogRunner('Invalid directory', message).run()
            return False

        # Make sure that it's possible to create/delete files
        # in the selected directory
        if not os.access(directory, os.F_OK | os.W_OK):
            message = ('Not enough <b>permissions</b> to write files '
                       'to selected directory:\n{0}\n\n'
                       'Unless permissions are changed accordingly '
                       'or checkbox-editor is executed '
                       'from a different user account with sufficient permissions, '
                       "it won't be possible to save changes to the same directory.\n\n"
                       'Would you like to continue?'
                       .format(directory))
            response = MessageDialogRunner('Insuficcient permissions',
                                           message,
                                           type=gtk.MESSAGE_WARNING,
                                           buttons=gtk.BUTTONS_YES_NO).run()
            if response != gtk.RESPONSE_YES:
                return False

        return True


    @handle_exceptions(InternalParsingError,
                       'Load operation <b>failed</b>')
    def load(self, directory):
        """
        Load data from directory
        """
        self.clear()
        self.problems = []

        self.directory = directory
        self.jobs_dir = os.path.join(directory, 'jobs')
        self.scripts_dir = os.path.join(directory, 'scripts')
        self.vc = VersionControl.get(directory,
                                     self.preferences.version_control.get('exclude'))

        if self.vc.has_changes():
            for widget_name in ['commit_menuitem', 'commit_toolbutton']:
                widget = self.builder.get_object(widget_name)
                widget.set_sensitive(True)

        if hasattr(self.vc, 'pull'):
            for widget_name in ['pull_menuitem', 'pull_toolbutton']:
                self.builder.get_object(widget_name).set_sensitive(True)

            if self.preferences.version_control.get('pull_on_open', False):
                pull_branch = self.preferences.version_control.get('pull_branch', None)
                if pull_branch:
                    pulled = self.vc.pull(pull_branch)
                    if not pulled:
                        return ExceptionHandled

        self._set_on_change_widgets()
        logging.info('Loading jobs: {0}'.format(self.directory))

        # Little trick that makes parsing work
        # when relative directories are in preferences
        os.chdir(directory)

        job_filename_pattern = re.compile('^[^.]+\.txt(\.in)?$')
        jobs_filenames = [os.path.join(self.jobs_dir, file)
                          for file in os.listdir(self.jobs_dir)
                          if job_filename_pattern.match(file)]

        parser_cache = {}
        row_cache = {}

        def safe_append(parent_iter, row_data):
            """
            Append row only if needed
            """
            filename = row_data[self.FILENAME]

            create_row = (not filename or filename not in row_cache
                          or not row_cache[filename].valid())

            if create_row:
                job_row = gtk.TreeStore.append(self, parent_iter, row_data)
                row_cache[filename] = TreeRowReference(self, self.get_path(job_row))
                return job_row

            logging.debug('{0!r} has been already parsed'.format(filename))
            cached_reference = row_cache[filename]
            cached_path = cached_reference.get_path()
            cached_iterator = self.get_iter(cached_path)
            cached_depth = self.iter_depth(cached_iterator)
            parent_depth = self.iter_depth(parent_iter) if parent_iter else 0

            # Create new row and remove the old one
            # if the new row is going to be deeper in the tree hierarchy
            if parent_depth >= cached_depth:
                # Warn user about cycles (jobs that include themselves)
                parent_path = self.get_path(parent_iter)
                if cached_path == parent_path[:cached_depth+1]:
                    parent_filename = self.get(parent_iter, self.FILENAME)[0]
                    parent_cached_iterator = self.iter_parent(cached_iterator)

                    message = ['A job in {0!r}'.format(parent_filename),
                               'includes {0!r}'.format(filename)]
                    if parent_cached_iterator:
                        parent_cached_filename = self.get(parent_cached_iterator,
                                                          self.FILENAME)[0]
                        message.append('which was already included by a job in {0!r}'
                                       .format(parent_cached_filename))
                    else:
                        message.append('which was already included as a root node')

                    problem = ('Cycle detected', '\n'.join(message))
                    if problem not in self.problems:
                        self.problems.append(problem)
                    return

                gtk.TreeStore.remove(self, cached_iterator)
                job_row = gtk.TreeStore.append(self, parent_iter, row_data)
                row_cache[filename] = TreeRowReference(self, self.get_path(job_row))
                return job_row


        def load_job_file(parent_iter, job_filename):
            """
            Load job description file using parent node in tree
            """
            if not os.path.isfile(job_filename):
                problem = ('File not found', job_filename)
                if not problem in self.problems:
                    self.problems.append(problem)
                return

            if job_filename in parser_cache:
                job_parser = parser_cache[job_filename]
            else:
                logging.info('Parsing jobs file: {0}'.format(job_filename))
                job_parser = DataParser(job_filename)
                parser_cache[job_filename] = job_parser


            for job_description in job_parser:
                if job_description.get('plugin') == 'local' and 'command' in job_description:
                    subjob_filenames = self.preferences.parse_job_command(job_description['command'],
                                                                          force=True)

                    if len(subjob_filenames) == 1:
                        subjob_filename = subjob_filenames[0]
                        job_row = safe_append(parent_iter, self._new_row(job_description['name'],
                                                                        job_description,
                                                                        subjob_filename))
                        if not job_row:
                            continue

                        load_job_file(job_row, subjob_filename)
                    else:
                        job_row = safe_append(parent_iter, self._new_row(job_description['name'],
                                                                        job_description))
                        for subjob_filename in self.preferences.parse_job_command(job_description['command']):
                            file_row = safe_append(job_row,
                                                   self._new_row(os.path.basename(subjob_filename),
                                                                 filename = subjob_filename))
                            if not job_row:
                                continue

                            load_job_file(file_row, subjob_filename)
                else:
                    job_row = gtk.TreeStore.append(self, parent_iter,
                                                   self._new_row(job_description['name'], job_description))


        # Traverse all job description files and parse them
        for job_filename in jobs_filenames:
            file_row = safe_append(None,
                                   self._new_row(os.path.basename(job_filename),
                                                 filename = job_filename))
            if not file_row:
                continue
            load_job_file(file_row, job_filename)


        # Mark the rows that have a duplicated name as invalid
        self._populate_name_cache()
        for name, row_references in self.name_cache.iteritems():
            if len(row_references) > 1:
                logging.error('{0!r} is being used as name by multiple jobs'
                              .format(name))

                for row_reference in row_references:
                    path = row_reference.get_path()
                    row=self[path]
                    self._set_problem_found(row.iter)

        # Mark the rows that have invalid dependencies set
        # (they're pointing to a name that doesn't exist)
        for row in self.get_all_rows():
            description = row[self.DESCRIPTION]
            if description:
                dependencies = description.get('depends')
                if dependencies:
                    for dependency in dependencies.split():
                        if dependency not in self.name_cache:
                            logging.warning('{0!r} dependency not found'
                                            .format(dependency))
                            self._set_problem_found(row.iter)

        # Warn about potential problems
        problems = self.validator.get_problems(self)
        if problems:
            glib.idle_add(self.validator.display_problems, problems, True)

        # Update revision number information
        vc = self.vc
        revno_label = self.builder.get_object('revno_label')
        if hasattr(vc, 'revno'):
            revno_label.set_text('rev: {0}'.format(vc.revno()))
        else:
            revno_label.set_text('')


    def _populate_name_cache(self):
        """
        Traverse all rows to add their name to the cache
        """
        self.name_cache.clear()
        for row in self.get_all_rows():
            description = row[self.DESCRIPTION]
            if description:
                name = description.get('name')
                if name:
                    row_reference = TreeRowReference(row.model, row.path)
                    self.name_cache[name].append(row_reference)


    def _set_on_change_widgets(self):
        """
        Set the names of the widgets that are subject to be activated
        on any change in the model based on the opened directory
        """
        on_change_widget_names = ['save_as_menuitem', 'revert_menuitem',]
        if os.access(self.directory, os.F_OK | os.W_OK):
            on_change_widget_names.extend(['save_menuitem', 'save_toolbutton'])
        self.on_change_widgets = [self.builder.get_object(widget_name)
                                  for widget_name in on_change_widget_names]


    @handle_exceptions((EnvironmentError, RenameOverwriteError),
                       'Save operation <b>failed</b>')
    def save(self):
        """
        Save changes to files
        """
        logging.info('Saving changes')

        success = self._rename_files()

        try:
            success = self._save_modified_children() and success
        except CancelException:
            success = False

        success = self._remove_files() and success

        if success:
            if (self.preferences.version_control.get('commit_on_save', False)
                and self.vc.commit()):
                for widget_name in ('push_menuitem', 'push_toolbutton'):
                    widget = self.builder.get_object(widget_name)
                    widget.set_sensitive(True)
            else:
                changes_found = self.vc.has_changes()

                for widget_name in ('commit_menuitem', 'commit_toolbutton'):
                    widget = self.builder.get_object(widget_name)
                    widget.set_sensitive(changes_found)

            # Disable save/revert options
            for widget in self.on_change_widgets:
                widget.set_sensitive(False)

        return success


    def _rename_files(self):
        """
        Perform pending rename operations
        """
        def rename_cmp(operation_1, operation_2):
            """
            Compare rename operations and detect conflicts
            """
            (old_name_1, new_name_1) = operation_1
            (old_name_2, new_name_2) = operation_2
            second_overwrites_first = old_name_1 == new_name_2
            first_overwrites_second = old_name_2 == new_name_1

            if second_overwrites_first and first_overwrites_second:
                raise RenameOverwriteError([os.path.relpath(fname, self.directory)
                                            for fname in operation_1],
                                           [os.path.relpath(fname, self.directory)
                                            for fname in operation_2])
            elif second_overwrites_first:
                return -1
            elif first_overwrites_second:
                return 1
            else:
                cmp(old_name_1, old_name_2)

        # Rename all files
        rename_operations = sorted(self.files_to_rename.items(), rename_cmp)

        for old_name, new_name in rename_operations:
            self.vc.rename(old_name, new_name)
            del self.files_to_rename[old_name]

            # Reset markers for renamed files
            for row in self.get_all_rows():
                if (row[self.FILENAME] == new_name
                    and not (row[self.MODIFIED]
                             or row[self.CHILDREN_MODIFIED]
                             or row[self.CHILDREN_REMOVED])):
                    row[self.FONT_DESCRIPTION] = pango.FontDescription()
                    break

            logging.debug('File renamed: {0} => {1}'
                          .format(old_name, new_name))
        return True


    def _save_modified_children(self, parent_row = None):
        """
        Look recursively for modified files and save them
        """
        success = True

        # Look for job description files that have been modified
        # Iterate over the root of the tree (self) if no parent row has been passed
        for job_row in [row for row in (parent_row.iterchildren() if parent_row else self)
                        if row[self.MODIFIED] or row[self.CHILDREN_MODIFIED]]:
            # Look for modified job description files further
            # in the tree hierarchy and save changes
            if job_row[self.CHILDREN_MODIFIED]:
                children_success = self._save_modified_children(job_row)
                if children_success:
                    job_row[self.CHILDREN_MODIFIED] = False
                success = success and children_success

            # Save job descriptions files that have been modified
            if job_row[self.MODIFIED]:
                jobs_metadata = [children_row[TreeStore.DESCRIPTION]
                                 for children_row in job_row.iterchildren()]

                filename = job_row[self.FILENAME]
                try:
                    self._save_metadata(filename, jobs_metadata)
                except EnvironmentError:
                    success = False

                    dialog = ExceptionDialog('Failed to save:\n'
                                             '<b>{0}</b>\n\n'
                                             'Would you like to continue?'
                                             .format(filename),
                                             buttons = (gtk.STOCK_NO, gtk.RESPONSE_NO,
                                                        gtk.STOCK_YES, gtk.RESPONSE_YES))
                    dialog.set_default_response(gtk.RESPONSE_YES)
                    dialog.show_all()
                    response = dialog.run()
                    dialog.destroy()
                    if response != gtk.RESPONSE_YES:
                        raise CancelException

                    continue
                else:
                    job_row[self.MODIFIED] = False


            # Reset font
            for children_row in job_row.iterchildren():
                if not (children_row[self.MODIFIED]
                        or children_row[self.CHILDREN_MODIFIED]
                        or children_row[self.CHILDREN_REMOVED]):
                    children_row[self.FONT_DESCRIPTION] = pango.FontDescription()

            if not (job_row[self.MODIFIED]
                    or job_row[self.CHILDREN_MODIFIED]
                    or job_row[self.CHILDREN_REMOVED]):
                job_row[self.FONT_DESCRIPTION] = pango.FontDescription()

        return success


    def _save_metadata(self, filename, elements_metadata):
        """
        Save metadata to a given file name
        """
        directory = os.path.dirname(filename)
        editable = (directory == self.directory
                    or directory.startswith(self.directory+os.path.sep))
        if not editable:
            logging.debug('{0!r} is not under currently opened directory. Skipping it.'
                          .format(filename))
            return

        i18n_file = True if filename.endswith('.in') else False
        encoder = DataEncoder.get('.txt', i18n_file)
        file_added = False if os.path.isfile(filename) else True
        with open(filename, 'w') as f:
            for element_metadata in elements_metadata:
                f.write(encoder.encode(element_metadata))

        if file_added:
            self.vc.add(filename)
            logging.debug('File added: {0}'.format(filename))

        logging.debug('Changes written to: {0}'.format(filename))


    def _remove_files(self):
        """
        Perform remove operations
        """
        # Remove all files that are no longer used
        files_removed = []
        for filename in self.files_to_remove:
            try:
                self.vc.remove(filename)
            except (EnvironmentError, BzrError):
                dialog = ExceptionDialog('Failed to remove:\n'
                                         '<b>{0}</b>\n\n'
                                         'Would you like to continue?'
                                         .format(filename),
                                         buttons = (gtk.STOCK_NO, gtk.RESPONSE_NO,
                                                    gtk.STOCK_YES, gtk.RESPONSE_YES))
                dialog.set_default_response(gtk.RESPONSE_YES)
                dialog.show_all()
                response = dialog.run()
                dialog.destroy()
                if response != gtk.RESPONSE_YES:
                    self.files_to_remove = [filename for filename in self.files_to_remove
                                            if filename not in files_removed]
                    return False

                continue
            else:
                files_removed.append(filename)
                self._unset_removed(filename)
                logging.debug('File removed: {0}'.format(filename))

        self.files_to_remove = [filename for filename in self.files_to_remove
                                if filename not in files_removed]

        return not bool(self.files_to_remove)


    def _unset_removed(self, filename, parent = None):
        """
        Remove filename from all file removed records
        """
        for row in (parent.iterchildren() if parent else self):

            try:
                index = row[self.CHILDREN_REMOVED].index(filename)
            except ValueError:
                continue

            self._unset_removed(filename, row)

            del row[self.CHILDREN_REMOVED][index]
            if not (row[self.MODIFIED]
                    or row[self.CHILDREN_MODIFIED]
                    or row[self.CHILDREN_REMOVED]):
                row[self.FONT_DESCRIPTION] = pango.FontDescription()


    def save_as(self, destination_directory):
        """
        Save changes to a different directory
        """
        source_directory = self.directory
        self.vc.clone(destination_directory)

        # Update all directory dependent variables in model
        logging.debug('Updating internal state to use new directory: {0!r}'
                      .format(destination_directory))
        self.directory = destination_directory
        self.jobs_dir = os.path.join(destination_directory, 'jobs')
        self.scripts_dir = os.path.join(destination_directory, 'scripts')
        self.vc = VersionControl.get(destination_directory,
                                     self.preferences.version_control['exclude'])
        def change_base_directory(filename):
            return os.path.join(destination_directory,
                                os.path.relpath(filename, source_directory))

        self.files_to_remove = [change_base_directory(filename)
                                for filename in self.files_to_remove]
        self.files_to_rename = dict([(change_base_directory(old_filename),
                                      change_base_directory(new_filename))
                                     for old_filename, new_filename
                                     in self.files_to_rename.items()])

        # Update all rows filenames
        for row in self.get_all_rows():
            filename = row[self.FILENAME]
            if filename:
                row[self.FILENAME] = change_base_directory(filename)

        self.save()
        self.preferences.directory = destination_directory
        self._set_on_change_widgets()


    def add_row(self, iterator):
        """
        Add a new row to the model
        """
        if iterator is None:
            filename = self.select_valid_job_file()
            if not filename:
                return

            name = os.path.basename(filename)
            new_row_iterator = self.append(iterator,
                                           self._new_row(name, filename = filename))
            logging.info('New file added')
        else:
            prefix = []
            row = self[iterator]

            while True:
                prefix.append(row[self.DISPLAY_NAME].rsplit('/', 1)[-1])
                row = row.parent
                if not row:
                    break

            prefix.reverse()
            prefix = '/'.join(prefix[1:])
            if prefix:
                prefix = '{0}/'.format(prefix)

            # Ensure that new job name is unique
            def job_name_generator(prefix=''):
                for count in itertools.count(1):
                    yield '{0}new_job_{1}'.format(prefix, count)

            for name in job_name_generator(prefix):
                same_name_rows = self.name_cache[name]
                if not same_name_rows:
                    break

            description = {'plugin': self.DEFAULT_PLUGIN,
                           'name': name}

            # Add default description field
            # if a value for that has been set in the validation file
            extra_validation_methods = getattr(self.preferences, 'extra_validation')
            if (extra_validation_methods
                and 'description_default' in extra_validation_methods):
                description['description'] = extra_validation_methods['description_default']()

            new_row_iterator = self.append(iterator,
                                           self._new_row(name, description))

            logging.info('New job added')
        self._set_modified(new_row_iterator)

        return new_row_iterator


    def append(self, parent, row_data):
        """
        Append row to the model with validation
        """
        iterator = gtk.TreeStore.append(self, parent, row_data)
        self._insert_row_helper(iterator)
        return iterator


    def insert_before(self, parent, sibling, row_data):
        """
        Insert before with validation
        """
        iterator = gtk.TreeStore.insert_before(self, parent, sibling, row_data)
        self._insert_row_helper(iterator)
        return iterator


    def insert_after(self, parent, sibling, row_data):
        """
        Insert after with validation
        """
        iterator = gtk.TreeStore.insert_after(self, parent, sibling, row_data)
        self._insert_row_helper(iterator)
        return iterator


    def _insert_row_helper(self, iterator):
        """
        Common tasks to all append/insert methods
        """
        # Update name cache
        row = self[iterator]
        name = row[self.DESCRIPTION].get('name')
        if name:
            reference = TreeRowReference(self, self.get_path(iterator))
            self.name_cache[name].append(reference)

        # Perform validation
        self._validate_row(iterator)

        # Set new row as modified
        self._set_modified(iterator)


    def _validate_row(self, iterator):
        """
        Validate row and take care of coloring
        """
        row = self[iterator]

        # Validate job description
        description = row[self.DESCRIPTION]
        invalid_field_names = self.validator.get_invalid_fields(description,
                                          getattr(self.preferences, 'extra_validation'))

        name = description.get('name')
        # Make sure that the same name isn't being used somewhere else
        if name:
            references = self.name_cache[name]
            if len(references) > 1:
                invalid_field_names.append('name')
                message = ('{0!r} is already being used as name by multiple jobs'
                           .format(name))
                logging.error(message)

                for reference in references:
                    same_name_row = self[reference.get_path()]
                    self._set_problem_found(same_name_row.iter)

        # Make sure that dependency is set pointing to a job that exists
        dependencies = description.get('depends')
        if dependencies:
            for dependency in dependencies.split():
                if dependency not in self.name_cache:
                    invalid_field_names.append('depends')
                    message =('{0!r} dependency not found'
                              .format(dependency))
                    logging.warning(message)
                    break

        # Take care of coloring in tree view
        if invalid_field_names:
            self._set_problem_found(iterator)
        else:
            self._set_problem_found(iterator, False)

        return iterator


    def select_valid_job_file(self, suggested_filename = None):
        """
        Select a valid job file
        That's a job file that isn't being used by another job
        """
        dialog = gtk.FileChooserDialog("Select a job description file",
                                       buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT,
                                                  gtk.STOCK_OPEN, gtk.RESPONSE_ACCEPT),
                                       action = gtk.FILE_CHOOSER_ACTION_SAVE)

        # Add filter
        filter = gtk.FileFilter()
        filter.set_name('Checkbox job files (*.txt, *.txt.in)')
        filter.add_pattern('*.txt')
        filter.add_pattern('*.txt.in')
        dialog.add_filter(filter)

        # Add shortcut folders
        for base_directory in set([os.path.realpath(d)
                                   for d in self.preferences.environment.values()]):
            directory = os.path.join(base_directory, 'jobs')
            if os.path.isdir(directory):
                dialog.add_shortcut_folder(directory)

        dialog.set_current_folder(self.jobs_dir)

        def parent_directories_iterator(filename):
            directory = os.path.dirname(filename)
            while directory:
                yield directory
                directory = os.path.dirname(directory)

        def set_suggested_filename():
            if suggested_filename:
                if os.path.isfile(suggested_filename):
                    dialog.set_filename(suggested_filename)
                else:
                    for directory in parent_directories_iterator(suggested_filename):
                        if os.path.isdir(directory):
                            dialog.set_current_folder(os.path.dirname(suggested_filename))
                            dialog.set_current_name(os.path.relpath(suggested_filename, directory))
                            break
                    else:
                        dialog.set_current_name(suggested_filename)
        set_suggested_filename()

        file_selected = False

        while not file_selected and dialog.run() == gtk.RESPONSE_ACCEPT:
            filename = dialog.get_filename()

            # It's assumed that suggested filename is valid
            # so no need to check against other jobs
            if filename == suggested_filename:
                file_selected = True
                break

            # Make sure that template extension is correct
            if not re.search('\.txt(\.in)?', filename):
                filename = '{0}.txt.in'.format(filename)

            # Make sure that filename isn't being used somewhere else
            file_rows = [file_row for file_row in self.get_all_rows()
                         if file_row[self.FILENAME]]
            for file_row in file_rows:
                if filename == file_row[self.FILENAME]:
                    job_display_name = file_row[self.DISPLAY_NAME]
                    logging.warning('Selected file is already being used by: {0}'
                                    .format(job_display_name))

                    error_dialog = gtk.MessageDialog(type=gtk.MESSAGE_ERROR,
                                               buttons=gtk.BUTTONS_OK)
                    error_dialog.set_title('Invalid job description file')
                    message = ('Job description file is already being used '
                               'by the following job: <b>{0}</b>.\n'
                               'The command field will remain unchanged.'
                               .format(job_display_name))
                    error_dialog.set_markup(message)
                    error_dialog.run()
                    error_dialog.destroy()

                    set_suggested_filename()
                    break
            else:
                file_selected = True

        dialog.destroy()
        return filename if file_selected else None


    def get_filename(self, row):
        """
        Return the filename in which the given job is included
        """
        if len(row.path) == 1:
            return row[self.FILENAME]

        row = row.parent
        filename = row[self.FILENAME]
        while filename is None:
            row = row.parent
            filename = row[self.FILENAME]
        return filename


    def remove(self, iterator):
        """
        Remove row from model
        """
        row = self[iterator]
        logging.debug('Removing job: {0}'.format(row[self.DISPLAY_NAME]))

        self._set_modified(iterator)

        filename = row[self.FILENAME]
        if filename:
            # Remember files to be removed
            directory = os.path.dirname(filename)
            opened_directory = self.directory
            is_under_opened_directory = (
                directory == opened_directory
                or directory.startswith(opened_directory+os.path.sep))
            if is_under_opened_directory:
                logging.debug('File will be removed: {0}'.format(filename))
                self.files_to_remove.append(filename)

            parent_row = row.parent
            while parent_row:
                parent_row[self.CHILDREN_REMOVED].append(filename)
                parent_row = parent_row.parent

        # Remove dependencies from other nodes
        display_name = row[self.DISPLAY_NAME]
        if (self.iter_depth(iterator) > 0
            and len(self.name_cache[display_name]) == 1):
            for child_row in self.get_all_rows():
                description = child_row[self.DESCRIPTION]
                if (description and 'depends' in description
                    and description['depends'] == display_name):
                    del description['depends']
                    self._set_modified(child_row.iter)

        # Remove all children if any
        for child_row in row.iterchildren():
            self.remove(child_row.iter)

        # Update name cache
        description = row[self.DESCRIPTION]
        row_reference = TreeRowReference(row.model, row.path)
        if description:
            name = description.get('name')
            if (name
                and name in self.name_cache
                and row_reference in self.name_cache[name]):

                self.name_cache[name].remove(row_reference)

                # Unmark as invalid another job with the same name
                if len(self.name_cache[name]) == 1:
                    same_name_reference = self.name_cache[name][0]
                    same_name_row = self[same_name_reference.get_path()]
                    if self.validator.is_valid(same_name_row[TreeStore.DESCRIPTION],
                                               getattr(self.preferences, 'extra_validation')):
                        self._set_problem_found(same_name_row.iter, False)

        # Unset any problem found before removing
        self._set_problem_found(row.iter, False)

        # Remove node
        gtk.TreeStore.remove(self, iterator)

        # Enable save/revert options
        for widget in self.on_change_widgets:
            widget.set_sensitive(True)


    def _new_row(self, name, description = None, filename = None,
                 modified = False, children_modified = False, children_removed = None,
                 font_description = None, markup=None, selected = False):
        """
        Create a tuple that can be appended to the model
        """
        markup = name
        if not description:
            description = None

        if not filename:
            filename = None

        if not children_removed:
            children_removed = []

        if not font_description:
            font_description = pango.FontDescription()

        return (name, description, filename,
                modified, children_modified, children_removed,
                font_description, markup, selected)


    def append_rename_operation(self, old_filename, new_filename):
        """
        Remember file rename operation
        """
        for key, value in self.files_to_rename.iteritems():
            if old_filename == value:
                old_filename = key
                break

        if old_filename:
            self.files_to_rename[old_filename] = new_filename
            logging.debug('File will be renamed: {0} => {1}'
                          .format(old_filename, new_filename))


    def _set_modified(self, iterator):
        """
        Set a row in a model as modified
        and mark all the parents in the tree as modified also
        """
        row = self[iterator]
        row[self.FONT_DESCRIPTION] = pango.FontDescription('bold italic')

        # The first parent linked to a file is set as modified
        # since it's the important thing to save later
        row = row.parent
        while row and not row[self.FILENAME]:
            row[self.FONT_DESCRIPTION] = pango.FontDescription('bold italic')
            row = row.parent

        if row and not row[self.MODIFIED]:
            row[self.MODIFIED] = True
            logging.debug('File modified: {0}'.format(row[self.FILENAME]))
            row[self.FONT_DESCRIPTION] = pango.FontDescription('bold italic')

            row = row.parent
            while row:
                if row[self.CHILDREN_MODIFIED]:
                    break
                row[self.CHILDREN_MODIFIED] = True
                row[self.FONT_DESCRIPTION] = pango.FontDescription('bold italic')
                row = row.parent

        # Enable save/revert options
        for widget in self.on_change_widgets:
            widget.set_sensitive(True)


    def _set_problem_found(self, iterator, problem_found=True):
        """
        Set a row in a model as error found (red foreground color)
        and higlight its ancestors (orange foreground color)
        """
        is_red_color_set = (re.compile('\<span foreground="red"\>.+\</span\>')
                            .match)
        is_orange_color_set = (re.compile('\<span foreground="orange"\>.+\</span\>')
                               .match)

        def any_child_has_problems(row):
            """
            Return true if any child has problems found
            """
            return any((is_red_color_set(child_row[self.MARKUP])
                        or is_orange_color_set(child_row[self.MARKUP]))
                       for child_row in row.iterchildren())

        row = self[iterator]
        if problem_found:
            # Set foreground color to red
            row[self.MARKUP] = ('<span foreground="red">{name}</span>'
                                .format(name=row[self.DISPLAY_NAME]))

            # Set ancestors foreground color to orange
            # until one of them is set to red
            # or the top of the tree is reached
            row = row.parent
            while row:
                if (is_red_color_set(row[self.MARKUP])
                    or is_orange_color_set(row[self.MARKUP])):
                    break

                row[self.MARKUP] = ('<span foreground="orange">{name}</span>'
                                    .format(name=row[self.DISPLAY_NAME]))
                row = row.parent
        else:
            # Remove foreground color
            # unless a child has any problem
            # (in that case set it to orange)
            if any_child_has_problems(row):
                row[self.MARKUP] = ('<span foreground="orange">{name}</span>'
                                    .format(name=row[self.DISPLAY_NAME]))
                row = row.parent
                return

            row[self.MARKUP] = row[self.DISPLAY_NAME]

            # Remove foreground color in ancestors
            # until one of them is set to red
            # or any other child is set to red
            # or the top of the tree is reached
            row = row.parent
            while row:
                if any_child_has_problems(row):
                    break

                if (is_red_color_set(row[self.MARKUP])
                    or not is_orange_color_set(row[self.MARKUP])):
                    break

                row[self.MARKUP] = row[self.DISPLAY_NAME]
                row = row.parent


    def set_selected(self, row, value):
        """
        Select/Deselect row
        If row is selected, then all children and ancestors are selected
        If row is deselected, then all children are deselected
        and ancestors if no other sibling is selected
        """
        def set_selected_ancestors(row, value):
            """
            Select/Deselect all ancestors of a given row
            """
            def get_ancestors(row):
                row = row.parent
                while row:
                    yield row
                    row = row.parent

            if value:
                # Select all ancestors
                ancestor_rows = itertools.takewhile(lambda row: not row[self.SELECTED],
                                                    get_ancestors(row))
                for ancestor_row in reversed(list(ancestor_rows)):
                    self._set_whitelist(ancestor_row, value)
                    ancestor_row[self.SELECTED] = value
            else:
                # Deselect all ancestors
                # if no other sibling is selected
                for parent_row in get_ancestors(row):
                    selected_child_rows = (child_row[self.SELECTED]
                                           for child_row
                                           in parent_row.iterchildren())
                    if any(selected_child_rows):
                        break

                    self._set_whitelist(parent_row, value)
                    parent_row[self.SELECTED] = value


        def set_selected_dependencies(row, value):
            """
            Select/Deselect based on dependencies row
            """
            description = row[self.DESCRIPTION]

            # Select job on which selected on depends
            if value:
                if description and 'depends' in description:
                    dependency_name = description['depends']

                    # TBD: Use a cache to find efficiently dependencies
                    # in any part of the tree. Not only for sibling jobs.
                    def get_dependencies(dependency_name, row):
                        for sibling_row in row.parent.iterchildren():
                            sibling_description = sibling_row[self.DESCRIPTION]

                            if ('name' in sibling_description
                                and sibling_description['name'] == dependency_name):
                                if 'depends' in sibling_description:
                                    for sibling_sibling_row \
                                            in get_dependencies(sibling_description['depends'],
                                                                sibling_row):
                                            yield sibling_sibling_row
                                yield sibling_row
                                break

                    for depends_row in get_dependencies(dependency_name, row):
                        self._set_whitelist(depends_row, value)
                        depends_row[self.SELECTED] = value

            # Deselect jobs that depend on the selected one
            elif description and 'name' in description:
                name = description['name']

                # TBD: Use a cache to find efficiently dependencies
                # in any part of the tree. Not only for sibling jobs.
                def deselect_dependency(name, row):
                    for sibling_row in row.parent.iterchildren():
                        sibling_description = sibling_row[self.DESCRIPTION]

                        if ('depends' in sibling_description
                            and sibling_description['depends'] == name):
                            self._set_whitelist(sibling_row, value)
                            sibling_row[self.SELECTED] = value

                            if 'name' in sibling_description:
                                deselect_dependency(sibling_description['name'],
                                                    sibling_row)

                deselect_dependency(name, row)


        def set_selected_children(parent, value):
            """
            Select/Deselect all children of a given row
            """
            for child_row in parent.iterchildren():
                self._set_whitelist(child_row, value)
                child_row[self.SELECTED] = value
                set_selected_children(child_row, value)


        if value:
            # Select ancestors and dependencies before row
            set_selected_ancestors(row, value)
            set_selected_dependencies(row, value)

            self._set_whitelist(row, value)
            row[self.SELECTED] = value
        else:
            self._set_whitelist(row, value)
            row[self.SELECTED] = value

            # Deselect ancestors and dependencies after row
            set_selected_dependencies(row, value)
            set_selected_ancestors(row, value)

        set_selected_children(row, value)

        # Set widgets sensitivity
        anything_selected = any(self.get_selected_rows())
        for widget_name in ['execute_menuitem',
                            'execute_toolbutton',
                            'save_whitelist_as_menuitem',
                            'export_whitelist_menuitem']:
            widget = self.builder.get_object(widget_name)
            widget.set_sensitive(anything_selected)


        if self.whitelist.filename:
            save_whitelist_menuitem = self.builder.get_object('save_whitelist_menuitem')
            save_whitelist_menuitem.set_sensitive(True)


    def _set_whitelist(self, row, value):
        """
        Add/Remove a job to/from the whitelist
        """
        description = row[self.DESCRIPTION]
        if description:
            name = description.get('name')
            if name:
                if value:
                    self.whitelist.append(name)
                else:
                    self.whitelist.remove(name)


    def get_all_columns(self, iter):
        """
        Return all columns for a given iterator
        """
        return self.get(iter, *range(self.get_n_columns()))


    def _copy_row_data(self, iterator):
        """
        Copy row data taking into account mutable objects
        """
        columns = self.get_all_columns(iterator)

        def copy_column(column):
            """
            Copy a single column
            """
            if isinstance(column, (list, dict)):
                return shallow_copy(column)

            if isinstance(column, pango.FontDescription):
                return pango.FontDescription(column.to_string())

            return column

        return [copy_column(column) for column in columns]


    def copy(self, origin_iter, destination_iter, insert_fn):
        """
        Copy recursively a row and all its children
        after/before/below another row
        """
        destination_columns = self._copy_row_data(origin_iter)
        inserted_iter = insert_fn(destination_iter, destination_columns)

        for child_row in self[origin_iter].iterchildren():
            self.copy_below(child_row.iter, inserted_iter)

        return inserted_iter


    def copy_below(self, origin_iter, destination_iter):
        """
        Copy recursively a row and all its children
        below another row
        """
        return self.copy(origin_iter, destination_iter, self.append)


    def copy_before(self, origin_iter, destination_iter):
        """
        Copy recursively a row and all its children
        before another row
        """
        insert_fn = lambda d, c: self.insert_before(None, d, c)
        return self.copy(origin_iter, destination_iter, insert_fn)


    def copy_after(self, origin_iter, destination_iter):
        """
        Copy recursively a row and all its children
        after another row
        """
        insert_fn = lambda d, c: self.insert_after(None, d, c)
        return self.copy(origin_iter, destination_iter, insert_fn)


class TreeRowReference(gtk.TreeRowReference):
    """
    A reference that compares to another one based on path
    """
    def __eq__(self, other):
        return self.get_path() == other.get_path()

