"""
preferences dialog module
"""
import pygtk
pygtk.require("2.0")
import gtk

import os, re, logging, pprint
import shlex
from glob import glob

from .util import MessageDialogRunner, CommandConfigParser as ConfigParser

class Preferences:
    """
    Preferences object
    that provides persistence through configuration file
    """
    ENV_VAR_REGEX = re.compile(r'\$(?:(\w+)|\{(\w+)\})')


    def __str__(self):
        return ('- General:\n{0}\n'
                '- Environment:\n{1}\n'
                '- Version control:\n{2}'
                .format(pprint.pformat(self.general),
                        pprint.pformat(self.environment),
                        pprint.pformat(self.version_control)))


    def load(self, directory = None):
        """
        Load preferences for a given directory
        """
        self.directory = directory
        self.problems = []
        self.extra_validation = None

        def get_config_filenames():
            """
            Return the name of all possible configuration filenames
            that should be checked
            """
            if (directory is not None
                and os.access(directory, os.F_OK | os.R_OK | os.W_OK | os.X_OK)):
                yield os.path.join(directory, '.checkbox-editor.cfg')

            yield os.path.expanduser('~/.checkbox-editor/checkbox-editor.cfg')

        config_filenames = list(get_config_filenames())

        # Set hardcoded default values
        self.general = {'checkbox_binary': '/usr/bin/checkbox-gtk',
                        'replace_tabs_with_spaces': True,
                        'number_of_spaces': 4,
                        'checkbox_data_cleanup': True,
                        'whitelist_resources': True,
                        }
        self.environment = {}
        if 'CHECKBOX_DATA' in os.environ:
            self.environment['CHECKBOX_DATA'] = os.environ['CHECKBOX_DATA']
        elif 'XDG_CACHE_HOME' in os.environ:
            self.environment['CHECKBOX_DATA'] = os.path.join(os.environ['XDG_CACHE_HOME'],
                                                             'checkbox')
        else:
            self.environment['CHECKBOX_DATA'] = os.path.expanduser('~/.cache/checkbox')
        self.version_control = {'exclude': []}

        # Overwrite environment with contents from file
        for config_filename in config_filenames:
            if not os.path.isfile(config_filename):
                continue

            logging.info(config_filenames)
            logging.info('Loading preferences from {0!r}...'.format(config_filename))
            self.config_filename = config_filename
            parser = ConfigParser()
            parser.read(config_filename)

            if parser.has_section('environment'):
                self.environment.update(dict([(name.upper(), path)
                                              for name, path in parser.items('environment')]))
            if parser.has_section('general'):
                general_options = parser.options('general')
                for option in ('checkbox_binary', 'extra_validation_file'):
                    if option in general_options:
                        self.general[option] = parser.get('general', option)
                for option in ('checkbox_data_cleanup', 'replace_tabs_with_spaces',
                               'whitelist_resources'):
                    if option in general_options:
                        self.general[option] = parser.getboolean('general', option)
                if 'number_of_spaces' in general_options:
                    self.general['number_of_spaces'] = parser.getint('general', 'number_of_spaces')

            if parser.has_section('version control'):
                version_control_options = parser.options('version control')
                for option in ('push_branch', 'pull_branch'):
                    if option in version_control_options:
                        self.version_control[option] = parser.get('version control', option)
                if 'exclude' in version_control_options:
                    self.version_control['exclude'] = parser.get('version control', 'exclude').split()
                for option in ('pull_on_open', 'commit_on_save', 'push_on_close'):
                    if option in version_control_options:
                        self.version_control[option] = parser.getboolean('version control', option)


            if ('extra_validation_file' in self.general
                and self.general['extra_validation_file']):
                extra_validation_file = self.general['extra_validation_file']
                if not os.path.isabs(extra_validation_file) and self.directory:
                    extra_validation_file = os.path.join(self.directory, extra_validation_file)

                if os.path.isfile(extra_validation_file):
                    extra_validation = {}
                    try:
                        execfile(extra_validation_file, extra_validation)
                    except:
                        logging.error('Problem while executing validation file: {0!r}'
                                      .format(extra_validation_file))
                    else:
                        self.extra_validation = extra_validation
                else:
                    logging.warning('Extra validation file not found: {0!r}'
                                    .format(extra_validation_file))

            logging.debug('Preferences loaded:\n'
                          '{0}'.format(self))
            break

        else:
            self.config_filename = config_filenames[0]
            logging.warning('Configuration file not found: {0}'.format(config_filenames))

        return self


    def update(self, general, environment, version_control):
        """
        Update preferences with new values
        return True if preferences are saved
        """
        update_values = []

        # Gather wich values must be updated
        keys = ['general', 'environment', 'version_control']
        for key in keys:
            value = locals()[key]
            if value != getattr(self, key):
                update_values.append(key)

        # Update values
        if update_values:
            self._save(general, environment, version_control)

            # Internal state is only updated if no exception was raised
            # when saving the new environment to a file
            for key in keys:
                value = locals()[key]
                setattr(self, key, value)

            logging.debug('Preferences updated:\n'
                          '{0}'.format(self))
            return update_values

        return []


    def _save(self, general, environment, version_control):
        """
        Save preferences to file
        """
        logging.info('Saving preferences to {0!r}...'.format(self.config_filename))

        # Create directories if needed
        if not os.path.exists(os.path.dirname(self.config_filename)):
            os.makedirs(os.path.dirname(self.config_filename))

        # Save preferences to file
        parser = ConfigParser()

        if general:
            parser.add_section('general')
            for name, value in general.iteritems():
                parser.set('general', name, value)

        if environment:
            parser.add_section('environment')
            for name, value in environment.iteritems():
                parser.set('environment', name, value)

        if version_control:
            parser.add_section('version control')
            for name, value in version_control.iteritems():
                # List of excluded files is stored
                # as filenames separated by a space
                if name == 'exclude':
                    value = ' '.join(value)
                parser.set('version control', name, value)

        with open(self.config_filename, 'w') as f:
            parser.write(f)


    def parse_job_command(self, command, force=False):
        """
        Parse job command and return the filenames that are embedded

        force: Used to return a filename for a pattern
        even when the file doesn't exist
        """
        logging.debug('Parsing command: {0}'.format(command))
        try:
            fragments = shlex.split(command)
        except ValueError:
            # Return nothing is shlex.split failed
            problem = ('Command parsing error', command)
            if problem not in self.problems:
                self.problems.append(problem)
            return []

        if not fragments[0] == 'cat':
            logging.warning("'cat' command not found: {0!r}".format(command))
            problem = ('Command parsing error', command)
            if problem not in self.problems:
                self.problems.append(problem)
            return []
        elif len(fragments) > 2:
            logging.warning("Unexpected command: {0!r}".format(command))
            problem = ('Command parsing error', command)
            if problem not in self.problems:
                self.problems.append(problem)
            return []


        # Iterate through all environment variables found in the path
        all_filenames = []
        for filename_pattern in self.expand(command).split()[1:]:
            filenames = [filename
                         for filename in glob(filename_pattern)
                         if not filename.endswith('~')]

            if not filenames and force:
                directory = os.path.dirname(filename_pattern)
                rootname, extension_pattern = os.path.splitext(os.path.basename(filename_pattern))

                if not self.ENV_VAR_REGEX.match(directory):
                    directory = os.path.realpath(directory)

                if extension_pattern == '.txt*':
                    if any(glob(os.path.join(directory, '*.txt.in'))):
                        extension = '.txt.in'
                    else:
                        extension = '.txt'

                    all_filenames.append(os.path.join(directory,
                                                      '{0}{1}'.format(rootname, extension)))

            for filename in filenames:
                if not self.ENV_VAR_REGEX.match(filename):
                    filename = os.path.realpath(filename)

                if not os.path.isfile(filename):
                    logging.error('File not found: (filename: {0}, command: {1}) '
                                  .format(filename, command))
                    continue
                all_filenames.append(filename)

        logging.debug('Filenames found in command: {0}'.format(all_filenames))
        return all_filenames


    def expand(self, line):
        """
        Expand environment variables in preferences in path
        """
        def expand_fragment(fragment):
            """
            Expand just one of the fragments in the line
            """
            for match in self.ENV_VAR_REGEX.finditer(fragment):
                varname = match.group(1) or match.group(2)
                replace_pattern = match.group(0)

                value = None
                # First look into preferences
                if varname in self.environment:
                    value = self.environment[varname]
                # Later look into shell environment
                elif varname in os.environ:
                    value = os.environ[varname]
                # Make sure that problem is reported later
                else:
                    problem = ('Environment variable not found',
                               varname)
                    if not problem in self.problems:
                        self.problems.append(problem)
                    logging.warning('Environment variable not found: {0} '
                                    '({1})'
                                    .format(varname, line))

                # Don't perform any default substitution
                # if file environment variable isn't found
                if value is not None:
                    fragment = fragment.replace(replace_pattern, value)

            return fragment

        # Iterate through all environment variables found in the path
        return ' '.join([expand_fragment(fragment)
                         for fragment in shlex.split(line)])


    def substitute(self, path):
        """
        Replace environment variables in preferences in path
        """
        # Sort environment variables by value length
        # this should guarantee longest replacement
        sorted_environment = sorted(self.environment.items(),
                                    lambda x, y: cmp(len(x[1]), len(y[1])),
                                    reverse=True)

        for name, value in sorted_environment:
            pattern = r'^%s((?=/)|$)' % os.path.realpath(value)
            replacement = '${%s}' % name
            path = re.sub(pattern, replacement, path)

        return path


class PreferencesDialog:
    """
    Wrapper class that loads widgets from glade file
    and runs dialog
    """
    GENERAL_WIDGET_NAMES = ('checkbox_binary', 'checkbox_data_cleanup',
                            'replace_tabs_with_spaces', 'number_of_spaces',
                            'whitelist_resources', 'extra_validation_file'
                            )
    VC_WIDGET_NAMES = ('pull_on_open', 'commit_on_save', 'push_on_close',
                       'pull_branch', 'push_branch',
                       )


    def __init__(self, preferences):
        """
        Create user interface widgets
        """
        self.preferences = preferences
        self.builder = gtk.Builder()
        ui_file = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                               'glade', 'preferences.glade')
        self.builder.add_from_file(ui_file)
        self.builder.connect_signals(self)

        self.dialog = self.builder.get_object('preferences_dialog')

        # Model for environment variables contains name/value pairs
        model = gtk.ListStore(str, str)

        # Setup treeview columns
        treeview = self.builder.get_object('environment_treeview')
        treeview.set_model(model)

        for index, column_name in enumerate(['Name', 'Value']):
            renderer = gtk.CellRendererText()
            column = gtk.TreeViewColumn(column_name, renderer,
                                        text=index)
            treeview.append_column(column)


    def run(self):
        """
        Run dialog
        """
        # Set dialog values from preferences
        environment_treeview = self.builder.get_object('environment_treeview')
        environment_model = environment_treeview.get_model()
        environment_model.clear()
        for row in self.preferences.environment.iteritems():
            environment_model.append(row)

        exclude_model = self.builder.get_object('exclude_liststore')
        if 'exclude' in self.preferences.version_control:
            for filename in self.preferences.version_control['exclude']:
                exclude_model.append((filename,))

        def set_widget_value(widget_name, value):
            """
            Set value to widget
            """
            widget = self.builder.get_object(widget_name)
            if isinstance(widget, gtk.CheckButton):
                return widget.set_active(value)
            if isinstance(widget, gtk.SpinButton):
                return widget.set_value(value)
            if isinstance(widget, gtk.Entry):
                return widget.set_text(value)
            if isinstance(widget, gtk.FileChooserButton):
                return widget.set_filename(value)

        for name, value in self.preferences.general.iteritems():
            set_widget_value(name, value)

        for name, value in self.preferences.version_control.iteritems():
            set_widget_value(name, value)

        response = self.dialog.run()
        updated_preferences = []

        if response == gtk.RESPONSE_OK:
            # Get preferences values from dialog
            def get_widget_value(widget_name):
                """
                Get value from widget
                """
                widget = self.builder.get_object(widget_name)
                if isinstance(widget, gtk.CheckButton):
                    return widget.get_active()
                if isinstance(widget, gtk.SpinButton):
                    return int(widget.get_value())
                if isinstance(widget, gtk.Entry):
                    return widget.get_text()
                if isinstance(widget, gtk.FileChooserButton):
                    value = widget.get_filename()

                    if value and self.preferences.directory:
                        value = os.path.relpath(widget.get_filename(),
                                                self.preferences.directory)
                    return value

            def empty_value(value):
                return value in (None, '')

            general = dict(filter(lambda t: not empty_value(t[1]),
                                  [(widget_name,
                                    get_widget_value(widget_name))
                                   for widget_name in self.GENERAL_WIDGET_NAMES]))

            environment = dict([(row[0], row[1])
                                for row in environment_model
                                if not empty_value(row[1])])

            version_control = dict(filter(lambda t: not empty_value(t[1]),
                                          [(widget_name,
                                            get_widget_value(widget_name))
                                           for widget_name in self.VC_WIDGET_NAMES]))
            exclude = [row[0] for row in exclude_model if row[0]]
            if exclude:
                version_control['exclude'] = exclude
            self.dialog.destroy()
            try:
                updated_preferences = self.preferences.update(general,
                                                              environment,
                                                              version_control)
            except IOError as exception:
                logging.error(exception)
                MessageDialogRunner('Error', str(exception)).run()
        else:
            self.dialog.destroy()

        return updated_preferences


    def replace_tabs_with_spaces_toggled_cb(self, button):
        """
        Set number_of_spaces sensitive according to replace_tabs_with_spaces
        """
        self.builder.get_object('number_of_spaces').set_sensitive(button.get_active())


    def add_environment_button_clicked_cb(self, button):
        """
        Add environment variable
        """
        dialog = self.builder.get_object('add_environment_dialog')
        dialog.set_title('Add preference')

        name_entry = self.builder.get_object('environment_name')
        name_entry.set_text('')
        directory_chooser = self.builder.get_object('environment_directory')
        directory_chooser.set_current_folder(os.getcwd())

        response = dialog.run()
        if response == gtk.RESPONSE_OK:
            name = name_entry.get_text()
            if name:
                name = name.upper()
                directory = directory_chooser.get_filename()
                treeview = self.builder.get_object('environment_treeview')
                model = treeview.get_model()
                iterator = model.append((name, directory))
                selection = treeview.get_selection()
                selection.select_iter(iterator)
        dialog.hide()
        return False


    def edit_environment_button_clicked_cb(self, button):
        """
        Edit environment variable
        """
        dialog = self.builder.get_object('add_environment_dialog')
        dialog.set_title('Edit preference')

        treeview = self.builder.get_object('environment_treeview')
        selection = treeview.get_selection()
        model, iterator = selection.get_selected()

        if iterator is None:
            return

        name_entry = self.builder.get_object('environment_name')
        name_entry.set_text(model.get_value(iterator, 0))
        directory_chooser = self.builder.get_object('environment_directory')
        directory_chooser.set_current_folder(model.get_value(iterator, 1))

        response = dialog.run()
        if response == gtk.RESPONSE_OK:
            name = name_entry.get_text()
            if name:
                name = name.upper()
                directory = directory_chooser.get_filename()
                model.set(iterator, 0, name, 1, directory)
        dialog.hide()
        return False


    def remove_environment_button_clicked_cb(self, button):
        """
        Remove selected environment variable
        """
        treeview = self.builder.get_object('environment_treeview')
        selection = treeview.get_selection()
        model, iterator = selection.get_selected()

        if iterator is not None:
            model.remove(iterator)

        return False


    def add_excluded_button_clicked_cb(self, button):
        """
        Add a filename to be excluded from commits
        """
        filename = self._select_excluded_filename()
        if filename:
            model = self.builder.get_object('exclude_liststore')
            model.append((filename,))

        return False


    def edit_excluded_button_clicked_cb(self, button):
        """
        Edit a filename to be excluded from commits
        """
        treeview = self.builder.get_object('exclude_treeview')
        selection = treeview.get_selection()
        model, iterator = selection.get_selected()

        if iterator:
            filename = model[iterator][0]
            filename = self._select_excluded_filename(filename)

            if filename:
                model[iterator] = (filename,)

        return False


    def _select_excluded_filename(self, initial_filename=None):
        """
        Select excluded filename
        """
        dialog = gtk.FileChooserDialog('Select a file',
                                       buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT,
                                                  gtk.STOCK_OK, gtk.RESPONSE_ACCEPT))

        directory = self.preferences.directory or os.getcwd()
        dialog.set_current_folder(directory)
        if initial_filename:
            dialog.set_filename(initial_filename)

        model = self.builder.get_object('exclude_liststore')
        filenames = [row[0] for row in model]

        while True:
            filename = None
            response = dialog.run()
            if response == gtk.RESPONSE_ACCEPT:
                abs_filename = dialog.get_filename()
                filename = os.path.relpath(dialog.get_filename(), directory)
                if filename.startswith(os.path.pardir):
                    log_message = ('Selected filename {0!r} must be '
                                   'under opened directory {1!r}'
                                   .format(abs_filename, directory))
                    logging.error(log_message)


                    dialog_message = ('Selected filename:\n{0}\n'
                                      'must be under opened directory:\n{1}'
                                      .format(abs_filename, directory))
                    MessageDialogRunner('Error', dialog_message).run()
                    continue

                if filename in filenames and filename != initial_filename:
                    log_message = ('Filename is already selected: {0!r}'
                                   .format(abs_filename))
                    logging.error(log_message)


                    MessageDialogRunner('Error',
                                        'Filename is already selected:\n{0}'
                                        .format(abs_filename)).run()
                    continue
            break

        dialog.destroy()
        return filename


    def remove_excluded_button_clicked_cb(self, button):
        """
        Remove a filename to be excluded from commits
        """
        treeview = self.builder.get_object('exclude_treeview')
        selection = treeview.get_selection()
        model, iterator = selection.get_selected()

        if iterator is not None:
            model.remove(iterator)

        return False


    def checkbox_binary_edit_button_clicked_cb(self, button):
        """
        Select checkbox binary
        """
        dialog = gtk.FileChooserDialog('Select a checkbox binary',
                                       buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT,
                                                  gtk.STOCK_OPEN, gtk.RESPONSE_ACCEPT),
                                       action = gtk.FILE_CHOOSER_ACTION_OPEN)

        checkbox_binary_entry = self.builder.get_object('checkbox_binary')
        checkbox_binary = checkbox_binary_entry.get_text().split()[0]
        dialog.set_filename(checkbox_binary)

        # Add shortcut folders
        for directory in self.preferences.environment.values():
            if os.path.isdir(directory):
                dialog.add_shortcut_folder(directory)

        response = dialog.run()

        if response == gtk.RESPONSE_ACCEPT:
            filename = dialog.get_filename()
            checkbox_binary_entry.set_text(filename)

        dialog.destroy()
        return False


    def destroy(self):
        """
        Destroy dialog
        """
        self.dialog.destroy()
