"""
Module to render templates to be used to create new projects
"""
import os, logging, platform, shutil, subprocess
from datetime import datetime

from jinja2 import Template


def get_available_templates():
    """
    Get the list of all available templates
    """
    templates_dir = os.getenv('CHECKBOX_EDITOR_TEMPLATES',
                              '/usr/share/checkbox-editor/templates')

    if not os.path.isdir(templates_dir):
        return []

    directories = filter(os.path.isdir,
                         [os.path.join(templates_dir, directory)
                          for directory in os.listdir(templates_dir)])
    return [DirectoryTemplate(d) for d in directories]


class DirectoryTemplate(object):
    TEMPLATE_EXTENSION = '.tmpl'

    def __init__(self, template_dir):
        self.template_dir = template_dir
        self.name = os.path.basename(template_dir)

    def render(self, target_dir):
        """
        Render a directory of files
        Filenames are always passed through the render engine
        but file contents it's only passed if extension matches
        """
        logging.debug('Rendering {0!r} template under {1!r}'
                      .format(self.name, target_dir))
        context = self._get_context(target_dir)
        environment = dict((var.upper(), value)
                           for var, value in context.items())
        environment.update({'DEBFULLNAME': context['maintainer_name'],
                            'DEBEMAIL': context['maintainer_email']})

        template_data_dir = os.path.join(self.template_dir, 'data')
        for root, directories, filenames in os.walk(template_data_dir):
            t_root = Template(os.path.relpath(root, template_data_dir)).render(**context)
            t_root = os.path.join(target_dir, t_root)

            # Write directories
            for directory in directories:
                t_dir = os.path.normpath(os.path.join(t_root,
                                                      Template(directory).render(**context)))
                full_directory = os.path.join(root, directory)
                if os.path.islink(full_directory):
                    os.symlink(os.readlink(full_directory), t_dir)
                    logging.debug('Directory link created: {0}'.format(t_dir))
                else:
                    os.mkdir(t_dir)
                    logging.debug('Directory created: {0}'.format(t_dir))

            # Write files
            for filename in filenames:
                logging.debug('Rendering {0!r}...'.format(filename))
                t_filename = os.path.join(t_root, Template(filename).render(**context))
                short_t_filename, extension = os.path.splitext(t_filename)
                full_filename = os.path.join(root,filename)
                if extension == self.TEMPLATE_EXTENSION:
                    with open(short_t_filename, 'w') as out_f:
                        with open(full_filename) as in_f:
                            out_f.write(Template(in_f.read()).render(**context))
                    shutil.copystat(full_filename, short_t_filename)
                    logging.debug('File created: {0}'.format(short_t_filename))
                else:
                    if os.path.islink(full_filename):
                        link_filename = os.readlink(full_filename)
                        short_link_filename, link_extension = os.path.splitext(link_filename)
                        if link_extension == self.TEMPLATE_EXTENSION:
                            os.symlink(short_link_filename, t_filename)
                        else:
                            os.symlink(link_filename, t_filename)
                        logging.debug('File link created: {0}'.format(t_filename))
                    else:
                        shutil.copyfile(full_filename, t_filename)
                        shutil.copystat(full_filename, t_filename)
                        logging.debug('File created: {0}'.format(t_filename))

        # Run postrender script
        postrender = os.path.join(self.template_dir, 'postrender')
        if os.path.isfile(postrender):
            logging.debug('Running postrender script...')
            process = subprocess.Popen(postrender,
                                       cwd=target_dir,
                                       env=environment)
            process.communicate()


        self._initialize_bzr_branch(target_dir)


    def _get_context(self, target_dir):
        """
        Create context to render individual template files
        """
        timezone = subprocess.Popen(('date', '+%z'),
                                     stdout=subprocess.PIPE).communicate()[0].rstrip()
        date = datetime.now().strftime('%a, %d %b %Y %H:%M:%S {0}'.format(timezone))
        year = datetime.now().strftime('%Y')
        distro = platform.dist()[2]
        fullname = os.getenv('DEBFULLNAME', 'OEM Community QA')
        email = os.getenv('DEBEMAIL', 'oem-community-qa@lists.launchpad.net')

        context = {'name': os.path.basename(target_dir),
                   'author_name': fullname,
                   'author_email': email,
                   'maintainer_name': fullname,
                   'maintainer_email': email,
                   'date': date,
                   'year': year,
                   'distro': distro}
        return context


    def _initialize_bzr_branch(self, target_dir):
        """
        Initialize bazaar branch
        """
        logging.info('Putting project under version control...')
        subprocess.Popen(('bzr', 'init', target_dir)).communicate()
        subprocess.Popen(('bzr', 'add', target_dir)).communicate()
        subprocess.Popen(('bzr', 'commit', target_dir, '--message=Automatic initial commit')).communicate()
