"""
treeview module
"""
import pygtk
pygtk.require("2.0")
import gtk
from .model import TreeStore, TreeRowReference
from .util import MessageDialogRunner

import os, re, logging


class Treeview(object):
    """
    Takes care of the treeview related events
    """
    ON_SELECTION_WIDGET_NAMES = ('add_toolbutton', 'add_menuitem',
                                 'remove_toolbutton', 'remove_menuitem',
                                 'cut_toolbutton', 'cut_menuitem',
                                 'copy_toolbutton', 'copy_menuitem')


    PASTE_WIDGET_NAMES = ('paste_toolbutton', 'paste_menuitem')

    PAGE_EMPTY = 0
    PAGE_FILE = 1
    PAGE_JOB = 2
    PAGE_WHITELIST = 3

    def __init__(self, builder, form, statusbar):
        self.builder = builder
        self.form = form
        self.statusbar = statusbar

        # Treestore using the same column types
        # as the one being displayed to be used as clipboard
        self.clipboard = gtk.TreeStore(*TreeStore.COLUMNS)

        self._init_popup_menus()


    def _init_popup_menus(self):
        """
        Initialize popup menus
        """
        treeview = self.builder.get_object('treeview')
        selection = treeview.get_selection()

        menu_data = {'file':
                         [('Append sibling', self.add_menuitem_activated_cb),
                          ('Append child', self.add_menuitem_activated_cb),
                          ('-', None),
                          ('Remove file', self.remove_menuitem_activated_cb),],
                     'job':
                         [('Append sibling', self.add_menuitem_activated_cb),
                          ('Append child', self.add_menuitem_activated_cb),
                          ('-', None),
                          ('Remove job', self.remove_menuitem_activated_cb),],
                     }

        self.menu = {}
        for menu_type, menus_data in menu_data.iteritems():
            menu = self.menu[menu_type] = gtk.Menu()

            for label, callback in menus_data:
                if label == '-':
                    menu_item = gtk.SeparatorMenuItem()
                else:
                    menu_item = gtk.MenuItem(label)

                # Set what to add/remove as a clue to the callback function
                menu_item.set_data('job_type', label.split()[-1])
                if callback:
                    menu_item.connect('activate', callback, selection)
                menu.append(menu_item)
            menu.attach_to_widget(treeview, None)
            menu.show_all()


    def selection_changed_cb(self, selection):
        """
        Update notebook with selected row information
        and set toolbuttons sensitivity
        """
        model, paths = selection.get_selected_rows()
        notebook = self.builder.get_object('notebook')

        if not paths:
            notebook.set_current_page(self.PAGE_EMPTY)

            for widget_name in self.ON_SELECTION_WIDGET_NAMES:
                self.builder.get_object(widget_name).set_sensitive(False)
            return False

        for widget_name in self.ON_SELECTION_WIDGET_NAMES:
            self.builder.get_object(widget_name).set_sensitive(True)

        row = model[paths[0]]
        description = row[TreeStore.DESCRIPTION]
        filename = row[TreeStore.FILENAME]

        # Make sure that command for 'local' jobs is edited properly
        local_plugin_set = (bool(description)
                            and 'plugin' in description and description['plugin'] == 'local')
        if local_plugin_set:
            self.builder.get_object('job_command').set_sensitive(False)
            self.builder.get_object('job_command_edit').set_sensitive(True)

            if not filename:
                for widget_name in ('add_toolbutton', 'add_menuitem'):
                    self.builder.get_object(widget_name).set_sensitive(False)
        else:
            self.builder.get_object('job_command').set_sensitive(True)
            self.builder.get_object('job_command_edit').set_sensitive(False)

        selection_log = []
        for index, path in enumerate(paths):
            row = model[path]
            selection_log.append(
                '{index}/{total}\n'
                '- display_name: {0!r}\n'
                '- description: {1!r}\n'
                '- filename: {2!r}\n'
                '- modified: {3}\n'
                '- children_modified: {4}\n'
                '- children_removed: {5}\n'
                '- selected: {6}'
                .format(row[TreeStore.DISPLAY_NAME],
                        row[TreeStore.DESCRIPTION],
                        row[TreeStore.FILENAME],
                        row[TreeStore.MODIFIED],
                        row[TreeStore.CHILDREN_MODIFIED],
                        row[TreeStore.CHILDREN_REMOVED],
                        row[TreeStore.SELECTED],
                        index=index+1, total=len(paths)))

        logging.debug('Selection changed ({0} rows):\n{1}'
                      .format(len(paths), '\n\n'.join(selection_log)))

        self._update_form_view()


    def _update_form_view(self):
        """
        Update form view according to treeview selection
        """
        treeview = self.builder.get_object('treeview')
        selection = treeview.get_selection()
        model, paths = selection.get_selected_rows()

        notebook = self.builder.get_object('notebook')
        whitelist_menuitem = self.builder.get_object('view_whitelist_menuitem')
        if whitelist_menuitem.get_active():
            notebook.set_current_page(self.PAGE_WHITELIST)
        elif paths:
            row = model[paths[0]]
            description = row[TreeStore.DESCRIPTION]

            # Only display job information when a description is available
            # Otherwise, displaye file information
            if not description:
                notebook.set_current_page(self.PAGE_FILE)
                filename_label = self.builder.get_object('filename_label')
                filename_label.set_text(row[TreeStore.DISPLAY_NAME])

                total, stats = model.get_statistics(row)
                file_statistics_label = self.builder.get_object('file_statistics_label')

                message = ['Jobs defined by plugin']
                for key in sorted(stats.keys()):
                    message.append('- {0}: <b>{1}</b>'.format(key, stats[key]))
                message.append('\n<big>Total: <b>{0}</b></big>'.format(total))
                message = '\n'.join(message)

                file_statistics_label.set_markup(message)
                return False

            # Fill data even it's not going to be diplayed
            # since view can change later from 'whitelist' to 'jobs'
            self.form.fill(row)
            notebook.set_current_page(self.PAGE_JOB)
        else:
            notebook.set_current_page(self.PAGE_EMPTY)

        return False


    def selection_changed_cb_after(self, selection):
        """
        Display popup menus if right mouse button was clicked
        """
        model, paths = selection.get_selected_rows()
        if not paths:
            return False

        path = paths[0]
        row = model[path]
        depth = model.iter_depth(row.iter)

        # Get right click event info and reset value
        event = selection.get_data('event')
        if event and hasattr(event, 'button'):
            selection.set_data('event', None)

            if depth == 0:
                self.menu['file'].popup(None, None, None, event.button, event.time)
            else:
                description = row[TreeStore.DESCRIPTION]

                # Take into account that jobs can be added
                # only to jobs that use local plugin
                menuitem=self.menu['job'].get_children()[1]
                if 'plugin' in description and description['plugin'] != 'local':
                    menuitem.set_sensitive(False)
                else:
                    menuitem.set_sensitive(True)
                self.menu['job'].popup(None, None, None, event.button, event.time)

        return False


    def add_toolbutton_clicked_cb(self, toolbutton):
        """
        Add a new job node
        """
        treeview = self.builder.get_object('treeview')
        selection = treeview.get_selection()

        # Behave as if the menuitem was activated
        self.add_menuitem_activated_cb(None, selection)
        return False


    def add_menuitem_activated_cb(self, menuitem, selection):
        """
        Add a new job node
        """
        model, paths = selection.get_selected_rows()

        path = paths[0]
        row = model[path]
        description = row[TreeStore.DESCRIPTION]

        # Select which row element to create
        # based on the job currently selected
        if (description is None
            or 'plugin' in description and description['plugin'] == 'local'):

            if menuitem and menuitem.get_data('job_type') == 'sibling':
                job_type = 'Sibling'
                parent_iterator = model.iter_parent(row.iter)
            else:
                job_type = 'Children'
                parent_iterator = row.iter

        else:
            job_type = 'Sibling'
            parent_iterator = model.iter_parent(row.iter)

        # Add & show the newly created row
        new_iterator = model.add_row(parent_iterator)
        if new_iterator:
            new_path = model.get_path(new_iterator)
            treeview = self.builder.get_object('treeview')
            treeview.expand_to_path(new_path)
            selection.unselect_all()
            selection.select_path(new_path)

            self.statusbar._write_tmp_message('{0} job created'.format(job_type))

        return False


    def remove_toolbutton_clicked_cb(self, toolbutton):
        """
        Remove file/suite/case item being selected
        """
        treeview = self.builder.get_object('treeview')
        selection = treeview.get_selection()

        # Behave as if the remove menuitem was activated
        self.remove_menuitem_activated_cb(None, selection)
        return False


    def remove_menuitem_activated_cb(self, menuitem, selection):
        """
        Remove file/suite/case item being selected
        """
        model, paths = selection.get_selected_rows()

        # Verify that all rows to remove
        # are under the opened directory
        invalid_job_selection_names = []
        opened_directory = model.directory
        for path in paths:
            row = model[path]
            filename = model.get_filename(row)
            directory = os.path.dirname(filename)
            can_be_removed = (directory == opened_directory
                              or directory.startswith(opened_directory+os.path.sep))
            if not can_be_removed:
                invalid_job_selection_names.append(model[path][TreeStore.DISPLAY_NAME])

        if invalid_job_selection_names:
            error_message = ("The following jobs <b>can't</b> be removed:\n{0}\n\n"
                             'Note: jobs can <b>only</b> be removed when they are included '
                             'in a file <b>under</b> the currently opened directory. '
                             .format('\n'.join(['- {0}'.format(name)
                                               for name in invalid_job_selection_names])))
            MessageDialogRunner('Error', error_message).run()
            return

        response = self._confirm_row_removal(model, paths)
        if response != gtk.RESPONSE_YES:
            return

        # Remove jobs references one by one
        # using references to verify
        # if they haven't been removed before
        # when removing any of their ancestors
        row_references = [TreeRowReference(model, path)
                          for path in paths]
        for row_reference in row_references:
            if row_reference.valid():
                model.remove(model[row_reference.get_path()].iter)

        # Set selection to parent of the first removed item
        path = paths[0]
        parent_path = path[:-1] if len(path) > 1 else (0,)
        selection.select_path(parent_path)

        self.statusbar._write_tmp_message('Jobs removed')

        return False


    def _confirm_row_removal(self, jobs_model, selected_paths):
        """
        Ask the user for confirmation before removing anything
        """
        dialog = gtk.Dialog('Remove row(s)',
                            self.builder.get_object('window'),
                            gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                            (gtk.STOCK_NO, gtk.RESPONSE_NO,
                             gtk.STOCK_YES, gtk.RESPONSE_YES))
        dialog.set_property('border-width', 10)
        dialog.set_default_response(gtk.RESPONSE_NO)
        dialog.vbox.set_spacing(5)
        hbox = gtk.HBox(spacing=5)
        image = gtk.image_new_from_stock(gtk.STOCK_DIALOG_WARNING,
                                         gtk.ICON_SIZE_DIALOG)
        image.set_alignment(0.5, 0.0)
        label = gtk.Label()
        label.set_markup('The following rows will be <b>removed</b>.\n'
                         'Would you like to continue?')
        hbox.pack_start(image, expand=False)
        hbox.pack_start(label, expand=False)
        dialog.vbox.pack_start(hbox, expand=False)

        store = gtk.ListStore(str)
        def add_recursive(row):
            store.append((row[jobs_model.DISPLAY_NAME],))
            for child_row in row.iterchildren():
                add_recursive(child_row)
        for path in selected_paths:
            row = jobs_model[path]
            add_recursive(row)
        treeview = gtk.TreeView(store)
        treeview.set_headers_visible(False)
        treeview.set_can_focus(False)
        selection = treeview.get_selection()
        selection.unselect_all()
        selection.set_mode(gtk.SELECTION_NONE)
        cell_renderer = gtk.CellRendererText()
        column = gtk.TreeViewColumn('Name', cell_renderer, text=0)
        treeview.append_column(column)

        scrolled_window = gtk.ScrolledWindow()
        scrolled_window.add(treeview)
        dialog.vbox.pack_start(scrolled_window, expand=True)

        dialog.show_all()

        response = dialog.run()
        dialog.destroy()
        return response


    def treeview_button_press_event_cb(self, treeview, event):
        """
        Show popup menu when right button is clicked
        """
        if (event.type == gtk.gdk.BUTTON_PRESS
            and event.button == 3):
            selection = treeview.get_selection()
            selection.set_data('event', event)

        return False


    def key_press_event_cb(self, widget, event):
        """
        Use some shortcuts to add/remove rows to the treeview
        """
        keyname = gtk.gdk.keyval_name(event.keyval)

        treeview = self.builder.get_object('treeview')
        selection = treeview.get_selection()
        model, paths = selection.get_selected_rows()
        if not paths:
            return

        if keyname == 'Delete':
            self.remove_menuitem_activated_cb(None, selection)
        elif keyname == 'Insert':
            self.add_menuitem_activated_cb(None, selection)


    def toggled_cb(self, renderer, path):
        """
        Set row selection properly
        """
        treeview = self.builder.get_object('treeview')
        model = treeview.get_model()
        iterator = model.get_iter(path)
        row = model[iterator]
        selected = row[model.SELECTED]

        model.set_selected(row, not selected)
        return False


    def row_edited_cb(self, renderer, path, new_name):
        """
        Set new display name for row
        """
        treeview = self.builder.get_object('treeview')
        model = treeview.get_model()
        iterator = model.get_iter(path)
        row = model[iterator]
        depth = model.iter_depth(row.iter)
        description = row[TreeStore.DESCRIPTION]
        name = row[TreeStore.DISPLAY_NAME]
        new_filename = os.path.join(model.jobs_dir, new_name)

        # Do nothing if name hasn't been modified
        if new_name == name:
            return

        # Validate filename
        if depth == 0 and not re.search('\.txt(\.in)?$', new_name):
            MessageDialogRunner('Invalid filename',
                                'Invalid filename: <b>{0}</b>\n'
                                'The extension should be either <b>.txt.in</b> or <b>.txt</b>'
                                .format(new_name)).run()
            return

        # Validate name
        new_name_exists = any(r for r in model.get_all_rows()
                              if (r[TreeStore.DISPLAY_NAME] == new_name
                                  and r.path != row.path))
        if new_name_exists:
            MessageDialogRunner('Invalid name',
                                '<b>{0}</b> is already being used in another row'
                                .format(new_name)).run()
            return

        # Set display name
        row[TreeStore.DISPLAY_NAME] = new_name
        logging.debug('name updated: {0}'.format(row[TreeStore.DISPLAY_NAME]))

        match = re.match('\<span foreground="(?P<color>\w+)"\>.+\</span\>',
                         row[TreeStore.MARKUP])
        if match:
            row[TreeStore.MARKUP] = ('<span foreground="{0}">{1}</span>'
                                     .format(match.group('color'), new_name))
        else:
            row[TreeStore.MARKUP] = new_name

        # Set filename
        if depth == 0:
            filename_label = self.builder.get_object('filename_label')
            filename_label.set_text(new_name)
            old_filename = row[TreeStore.FILENAME]
            model.append_rename_operation(old_filename, new_filename)

            row[TreeStore.FILENAME] = new_filename
            logging.debug('filename updated: {0}'.format(row[TreeStore.FILENAME]))
        else:
            # Update name cache
            row_reference = TreeRowReference(row.model, row.path)
            if (name
                and name in model.name_cache
                and row_reference in model.name_cache[name]):
                model.name_cache[name].remove(row_reference)
            if new_name:
                model.name_cache[new_name].append(row_reference)

        # Set name in description and in widget
        if description:
            description['name'] = new_name
            widget = self.builder.get_object('job_name')
            widget.set_text(new_name)
            logging.debug('description updated: {0}'.format(description))

        # Mark row as modified
        model._set_modified(iterator)


    def drag_data_received_cb(self, treeview, context, x, y, selection, info, timestamp):
        """
        Copy/Move a job subtree to a differenct location
        """
        selection = treeview.get_selection()
        model, paths = selection.get_selected_rows()
        origin_rows = [model[path] for path in paths]
        opened_directory = model.directory

        drop_info = treeview.get_dest_row_at_pos(x, y)
        if not drop_info:
            context.finish(False, False, timestamp)
            return
        path, position = drop_info
        destination_row = model[path]

        # Find out whether destination is under currently opened directory or not
        if (position in (gtk.TREE_VIEW_DROP_BEFORE, gtk.TREE_VIEW_DROP_AFTER)
            and len(destination_row.path) > 1):
            row = destination_row.parent
        else:
            row = destination_row
        destination_filename = model.get_filename(row)
        destination_directory = os.path.dirname(destination_filename)
        destination_editable = (destination_directory == opened_directory
                                or destination_directory.startswith(opened_directory+os.path.sep))
        destination_description = destination_row[TreeStore.DESCRIPTION]
        destination_may_have_children = (destination_description is None
                                         or destination_description.get('plugin') == 'local')

        # Verify that every row selection is correct
        errors = []

        if not destination_editable:
            errors.append("Jobs <b>can't</b> be moved/copied to <b>{0}</b>, "
                          "since it's defined in a file not located "
                          'under the currently opened directory'
                          .format(destination_row[model.DISPLAY_NAME]))

        for origin_row in origin_rows:
            if (len(origin_row.path) <= 1
                and (len(destination_row.path) > 1
                     or position in (gtk.TREE_VIEW_DROP_INTO_OR_BEFORE,
                                     gtk.TREE_VIEW_DROP_INTO_OR_AFTER))):
                errors.append('<b>{0}</b> can <b>only</b> be moved/copied '
                              'to other position at the same level'
                              .format(origin_row[model.DISPLAY_NAME]))
            elif (len(origin_row.path) > 1 and len(destination_row.path) <= 1
                and position in (gtk.TREE_VIEW_DROP_BEFORE,
                                 gtk.TREE_VIEW_DROP_AFTER)):
                errors.append("<b>{0} can't</b> be moved/copied "
                              'to the root of the tree'
                              .format(origin_row[model.DISPLAY_NAME]))
            elif ((origin_row.path == destination_row.path
                   and position in (gtk.TREE_VIEW_DROP_INTO_OR_BEFORE,
                                    gtk.TREE_VIEW_DROP_INTO_OR_AFTER))
                  or (origin_row.path != destination_row.path
                      and origin_row.path == destination_row.path[:len(origin_row.path)])):
                errors.append("<b>{0} can't</b> be moved/copied "
                              'to a position under itself'
                              .format(origin_row[model.DISPLAY_NAME]))

        # If errors have been found, then display message
        # and cancel drag&drop operation
        if errors:
            error_message = ('The following problems were found:\n\n{0}'
                             .format('\n\n'.join(['- {0}'.format(error) for error in errors])))
            MessageDialogRunner('Drag & drop failure', error_message).run()
            context.finish(False, False, timestamp)
            return

        pasted_references = []
        for origin_row in origin_rows:
            # If origin and destination rows are siblings
            # just move the rows in the model
            siblings = (len(origin_row.path) == len(destination_row.path)
                        and origin_row.path[:-1] == destination_row.path[:-1])
            if siblings:
                if (position == gtk.TREE_VIEW_DROP_BEFORE
                    or (position == gtk.TREE_VIEW_DROP_INTO_OR_BEFORE
                        and not destination_may_have_children)):
                    pasted_references.append(TreeRowReference(model, origin_row.path))
                    model.move_before(origin_row.iter, destination_row.iter)
                    continue

                if (position == gtk.TREE_VIEW_DROP_AFTER
                    or (position == gtk.TREE_VIEW_DROP_INTO_OR_AFTER
                        and not destination_may_have_children)):
                    pasted_references.append(TreeRowReference(model, origin_row.path))
                    model.move_after(origin_row.iter, destination_row.iter)
                    continue

            # When they aren't siblings, copy the rows to the destination location
            # and, after that, remove the rows when the origin row is editable,
            # that is, can be removed
            origin_filename = model.get_filename(origin_row)
            origin_directory = os.path.dirname(origin_filename)
            origin_editable = (origin_directory == opened_directory
                               or origin_directory.startswith(opened_directory+os.path.sep))

            if position == gtk.TREE_VIEW_DROP_BEFORE:
                copy_method = model.copy_before
            elif gtk.TREE_VIEW_DROP_INTO_OR_BEFORE:
                if destination_may_have_children:
                    copy_method = model.copy_below
                else:
                    copy_method = model.copy_before
            elif position == gtk.TREE_VIEW_DROP_AFTER:
                copy_method = model.copy_before
            elif position == gtk.TREE_VIEW_DROP_INTO_OR_AFTER:
                if destination_may_have_children:
                    copy_method = model.copy_below
                else:
                    copy_method = model.copy_after

            iterator = copy_method(origin_row.iter, destination_row.iter)
            pasted_references.append(TreeRowReference(model,
                                                      model.get_path(iterator)))
            if origin_editable:
                model.remove(origin_row.iter)


        # Select pasted paths
        selection.unselect_all()
        for pasted_reference in pasted_references:
            pasted_path = pasted_reference.get_path()
            treeview.expand_to_path(pasted_path)
            treeview.collapse_row(pasted_path)
            selection.select_path(pasted_path)

        context.finish(True, False, timestamp)


    def cut_menuitem_activate_cb(self, widget):
        """
        Cut selected job(s)
        """
        treeview = self.builder.get_object('treeview')
        selection = treeview.get_selection()
        model, paths = selection.get_selected_rows()
        references = [TreeRowReference(model, path)
                      for path in paths]

        invalid_job_names = []
        for path in paths:
            row = model[path]
            destination_filename = model.get_filename(row)
            destination_directory = os.path.dirname(destination_filename)
            is_destination_valid = (destination_directory == model.directory
                                    or destination_directory.startswith(model.directory+os.path.sep))
            if not is_destination_valid:
                invalid_job_names.append(row[TreeStore.DISPLAY_NAME])

        if invalid_job_names:
            error_message = ("The following jobs <b>can't</b> be cut:\n{0}\n\n"
                             'Note: jobs can <b>only</b> be cut when they are included '
                             'in a file <b>under</b> the currently opened directory. '
                             .format('\n'.join(['- {0}'.format(name)
                                               for name in invalid_job_names])))
            MessageDialogRunner('Error', error_message).run()
            return


        self.clipboard.clear()
        for path in paths:
            self._append_to_clipboard(model[path])

        for reference in references:
            model.remove(model.get_iter(reference.get_path()))

        for widget_name in self.PASTE_WIDGET_NAMES:
            widget = self.builder.get_object(widget_name)
            widget.set_sensitive(True)

        logging.debug('{0} rows cut to clipboard'
                      .format(len(self.clipboard)))


    def copy_menuitem_activate_cb(self, widget):
        """
        Copy selected job(s)
        """
        treeview = self.builder.get_object('treeview')
        selection = treeview.get_selection()
        model, paths = selection.get_selected_rows()

        self.clipboard.clear()
        for path in paths:
            self._append_to_clipboard(model[path])

        for widget_name in self.PASTE_WIDGET_NAMES:
            widget = self.builder.get_object(widget_name)
            widget.set_sensitive(True)

        logging.debug('{0} rows copied to clipboard'
                      .format(len(self.clipboard)))


    def _append_to_clipboard(self, row, parent_iterator=None):
        """
        Copy rows to clipboard
        """
        model = row.model
        row_data = model._copy_row_data(row.iter)
        new_iterator = self.clipboard.append(parent_iterator, row_data)

        for child_row in row.iterchildren():
            self._append_to_clipboard(child_row, new_iterator)


    def paste_menuitem_activate_cb(self, widget):
        """
        Paste cut/copied job(s)
        """
        treeview = self.builder.get_object('treeview')
        selection = treeview.get_selection()
        model, paths = selection.get_selected_rows()

        if len(paths) == 0:
            error_message = ('Please <b>select</b> where the jobs should be pasted\n'
                             'in the <b>tree view</b>')
            MessageDialogRunner('Error', error_message).run()
            return

        invalid_job_names = []
        for path in paths:
            destination_row = model[path]

            # If selected row isn't a suite,
            # paste as a sibling job instead of as child job
            description = destination_row[model.DESCRIPTION]
            if description:
                plugin = description.get('plugin')
                if plugin != 'local':
                    destination_row = destination_row.parent

            destination_filename = model.get_filename(destination_row)
            destination_directory = os.path.dirname(destination_filename)
            is_destination_valid = (destination_directory == model.directory
                                    or destination_directory.startswith(model.directory+os.path.sep))

            if not is_destination_valid:
                invalid_job_names.append(destination_row[TreeStore.DISPLAY_NAME])

        if invalid_job_names:
            error_message = ("Jobs <b>can't</b> be pasted to the following locations:\n{0}\n\n"
                             'Note: jobs can <b>only</b> be pasted to a locations which is included '
                             'in a file <b>under</b> the currently opened directory. '
                             .format('\n'.join(['- {0}'.format(name)
                                               for name in invalid_job_names])))
            MessageDialogRunner('Error', error_message).run()
            return

        def paste_row_data(iterator, parent_iterator=None):
            row_data = self.clipboard.get(iterator,
                                          *range(self.clipboard.get_n_columns()))
            new_iterator = model.append(parent_iterator, row_data)

            for child_row in self.clipboard[iterator].iterchildren():
                paste_row_data(child_row.iter, new_iterator)

            return model.get_path(new_iterator)

        pasted_paths = []
        for path in paths:
            destination_iterator = model.get_iter(path)
            pasted_paths.extend([paste_row_data(origin_row.iter,
                                                destination_iterator)
                                 for origin_row in self.clipboard])

        selection.unselect_all()
        for pasted_path in pasted_paths:
            treeview.expand_to_path(pasted_path)
            treeview.collapse_row(pasted_path)
            selection.select_path(pasted_path)

        logging.debug('{0} rows pasted from clipboard'.format(len(self.clipboard)))


    def view_menuitem_toggled_cb(self, menuitem):
        """
        Update notebook page when view menuitem state changes
        """
        self._update_form_view()
