"""
Utiliy classes
"""
import pygtk
pygtk.require("2.0")
import gtk, glib

import sys, os, re, logging, subprocess, threading
from ConfigParser import ConfigParser
from functools import wraps
from traceback import format_exception, extract_tb

class CommandLauncher(threading.Thread):
    """
    Execute a command and display results in a dialog
    """
    def __init__(self, command_str, env=None):
        threading.Thread.__init__(self)

        # Application will be able to exit
        # even when subprocess hasn't finished
        self.daemon = True

        self.command_str = command_str
        self.env = env

        self.process = None
        self.stdout = None
        self.stderr = None


    def run(self):
        """
        Task to be executed in a separated thread
        """
        process = subprocess.Popen(self.command_str,
                                   shell=True,
                                   env=self.env,
                                   stdout = subprocess.PIPE,
                                   stderr = subprocess.PIPE)
        stdout, stderr = process.communicate()

        logging.debug('Command executed: '
                      '(command_str: {0}, '
                      'status: {1}, '
                      'stdout: {2}, '
                      'stderr: {3})'
                      .format(self.command_str,
                              process.returncode,
                              repr(stdout),
                              repr(stderr)))

        self.process = process
        self.stdout = stdout
        self.stderr = stderr
        glib.idle_add(self._display)


    def _display(self):
        """
        Display results in custom dialog
        """
        # Create custom dialog
        # to display possibly long results in a scrolled window
        dialog = gtk.Dialog('Command output',
                            buttons = (gtk.STOCK_OK, gtk.RESPONSE_OK))
        dialog.set_icon_name('checkbox-editor')
        dialog.set_default_response(gtk.RESPONSE_OK)
        dialog.set_property('border-width', 10)
        dialog.vbox.set_spacing(5)
        hbox = gtk.HBox(spacing=5)
        image = gtk.image_new_from_stock((gtk.STOCK_DIALOG_INFO
                                          if self.process.returncode == 0
                                          else gtk.STOCK_DIALOG_ERROR),
                                         gtk.ICON_SIZE_DIALOG)
        image.set_alignment(0.5, 0.0)
        label = gtk.Label()
        label.set_markup('Command execution <b>{0}</b>'
                         .format('success'
                                 if self.process.returncode == 0
                                 else 'failure'))
        hbox.pack_start(image, expand=False)
        hbox.pack_start(label, expand=False)
        dialog.vbox.pack_start(hbox, expand=False)

        table_data = [('Command', self.command_str),
                      ('Status', str(self.process.returncode))]
        if self.stdout:
            table_data.append(('Standard output', self.stdout))
        if self.stderr:
            table_data.append(('Standard error', self.stderr))

        table = gtk.Table(rows=len(table_data),
                          columns=2, homogeneous=False)
        table.set_row_spacings(5)
        table.set_col_spacings(5)
        for row_number, (label_str, data_str) in enumerate(table_data):
            if '\n' in data_str:
                expander = gtk.Expander(label_str)
                scrolled_window = gtk.ScrolledWindow()
                scrolled_window.set_policy(gtk.POLICY_AUTOMATIC,
                                           gtk.POLICY_AUTOMATIC)
                textview = gtk.TextView()
                textview.set_editable(False)
                buffer = textview.get_buffer()
                buffer.set_text(data_str)
                scrolled_window.add(textview)
                expander.add(scrolled_window)
                table.attach(expander, 0, 2, row_number, row_number+1)
            else:
                label = gtk.Label('{0}:'.format(label_str))
                label.set_alignment(1.0, 0.5)
                table.attach(label, 0, 1, row_number, row_number+1, gtk.FILL, gtk.FILL)

                if len(data_str) > 40:
                    entry = gtk.Entry()
                    entry.set_text(data_str)
                    entry.set_property('editable', False)
                    table.attach(entry, 1, 2, row_number, row_number+1, yoptions=gtk.FILL)
                else:
                    label = gtk.Label(data_str)
                    label.set_alignment(0.0, 0.5)
                    label.set_selectable(True)
                    table.attach(label, 1, 2, row_number, row_number+1, yoptions=gtk.FILL)

        dialog.vbox.pack_start(table)
        dialog.connect('response', lambda w, d: dialog.destroy())

        # Make ok button grab focus
        # so that enter key defaults to clicking in the button
        ok_button = dialog.get_action_area().get_children()[0]
        ok_button.grab_focus()

        dialog.show_all()


class SignalBlocker:
    def __init__(self, widget, callback):
        self.widget = widget
        self.callback = callback

    def __enter__(self):
        self.widget.handler_block_by_func(self.callback)

    def __exit__(self, type, value, traceback):
        self.widget.handler_unblock_by_func(self.callback)
        return False


class GtkCriticalSection:
    def __enter__(self):
        gtk.gdk.threads_enter()

    def __exit__(self, type, value, traceback):
        gtk.gdk.threads_leave()
        return False


class ExceptionHandled:
    pass


class IgnoreException(Exception):
    pass


def handle_exceptions(exceptions_handled, markup_message = None):
    """
    Catch exceptions, print them to log and display dialog with information also
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except exceptions_handled:
                type, value, traceback = sys.exc_info()
                exception_handler(type, value, traceback, markup_message)
                return ExceptionHandled

        return wrapper
    return decorator


def report_bug():
    """
    Report bug using apport
    """
    module = {}
    APPORT_GTK = '/usr/share/apport/apport-gtk'
    execfile(APPORT_GTK, module)
    sys.argv = [APPORT_GTK]
    ui = module['GTKUserInterface']()
    ui.options.package = 'checkbox-editor'

    # Get current instance of Editor
    from .editor import Editor
    editor = Editor.get()
    ui.options.log_filename = editor.options.log_file
    ui.options.cfg_filename = editor.preferences.config_filename

    ui.run_report_bug()


def exception_handler(type, value, traceback, markup_message = None):
    """
    Log and display all exceptions (thread-aware)
    """
    def handle_response(dialog, response):
        """
        Report bug through apport
        """
        dialog.destroy()

        if response == ExceptionDialog.REPORT_BUG:
            report_bug()

    def display_dialog():
        """
        Display custom dialog with exception information
        """
        with GtkCriticalSection():
            dialog = ExceptionDialog(markup_message,
                                     exc_data = (type, value, traceback))
            dialog.connect('response', handle_response)
            dialog.show_all()

    # Display dialog later in the main thread
    glib.idle_add(display_dialog)


class ExceptionDialog(gtk.Dialog):
    """
    Dialog that shows information regarding an exception
    """
    CONTINUE = 1
    REPORT_BUG = 2

    def __init__(self, markup_message = None, buttons = None, exc_data = None):
        # Log exception data before creating the dialog
        if exc_data is None:
            exc_data = sys.exc_info()
        exception_str = ''.join(format_exception(*exc_data))
        logging.error(exception_str)

        gtk.Dialog.__init__(self, 'Error',
                            flags = gtk.DIALOG_MODAL,
                            buttons = buttons)
        self.set_property('border-width', 10)
        self.vbox.set_spacing(5)

        if not buttons:
            self.add_button('Continue', self.CONTINUE)
            button = gtk.Button('Report problem...')
            image = gtk.image_new_from_icon_name('apport',
                                                 gtk.ICON_SIZE_BUTTON)
            button.set_image(image)
            self.add_action_widget(button, self.REPORT_BUG)

        self.set_default_response(self.CONTINUE)

        hbox = gtk.HBox(spacing=5)
        image = gtk.image_new_from_stock(gtk.STOCK_DIALOG_ERROR,
                                         gtk.ICON_SIZE_DIALOG)
        image.set_alignment(0.5, 0.0)
        label = gtk.Label()
        if markup_message:
            label.set_markup(markup_message)
        else:
            _type, _value, traceback = exc_data
            (filename, line_number, function_name, text) = extract_tb(traceback)[-1]
            label.set_markup('Unexpected exception in:\n'
                             '- File: <b>{0}</b>, line <b>{1}</b>\n'
                             '- Function: <b>{2}</b>'
                             .format(os.path.basename(filename),
                                     line_number,
                                     function_name))
        hbox.pack_start(image, expand=False)
        hbox.pack_start(label, expand=False)
        self.vbox.pack_start(hbox, expand=False)

        expander = gtk.expander_new_with_mnemonic('_More information')

        scrolled_window = gtk.ScrolledWindow()
        textview = gtk.TextView()
        textview.set_editable(False)
        buffer = textview.get_buffer()
        buffer.set_text(exception_str)
        scrolled_window.add(textview)

        expander.add(scrolled_window)
        self.vbox.pack_start(expander)


class CommandConfigParser(ConfigParser):
    """
    Config parser that provides extra interpolation
    for shell command substitution
    """
    COMMAND_REGEX = re.compile(r'\$\(([^)]+)\)')


    def _interpolate(self, *args, **kwargs):
        """
        Interpolate string
        """
        value = ConfigParser._interpolate(self, *args, **kwargs)
        value = self.COMMAND_REGEX.sub(self._interpolate_cmd, value)
        return value


    def _interpolate_cmd(self, match):
        """
        Run command and return first line from standard output
        """
        cmd = match.group(1)
        process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
        stdout, _ = process.communicate()
        stdout = stdout.splitlines()[0]
        return stdout


class MessageDialogRunner(object):
    """
    Create a MessageDialog, run it, destroy it and return response
    """
    def __init__(self, title, message,
                 type=gtk.MESSAGE_ERROR,
                 buttons=gtk.BUTTONS_OK,
                 flags=gtk.DIALOG_MODAL,
                 default_response=None):
        """
        Create MessageDialog object
        """
        self.title = title
        self.message = message
        self.type = type
        self.buttons = buttons
        self.flags = flags
        self.default_response = default_response


    def run(self):
        """
        Run dialog, destroy it and return response
        """
        dialog = gtk.MessageDialog(type=self.type,
                                   buttons=self.buttons,
                                   flags=self.flags)
        dialog.set_title(self.title)
        dialog.set_markup(self.message)
        if self.default_response:
            dialog.set_default_response(self.default_response)
        response = dialog.run()
        dialog.destroy()
        return response
