"""
Validation classes
"""
import pygtk
pygtk.require("2.0")
import gtk

import logging, itertools, string

class JobValidator(object):
    """Validate job"""

    MANDATORY_FIELDS = ('name',)
    MANDATORY_PLUGIN_FIELDS = {
        'local': ('command', 'description'),
        'remote': ('command', 'description'),
        'manual': ('description',),
        'shell': ('command', 'description'),
        'attachment': ('command',),
        'resource': ('command',),
        }


    def is_valid(self, job, extra_validation_methods=None):
        """
        Check that all mandatory fields contain some data
        """
        return not self.get_invalid_fields(job, extra_validation_methods)


    def get_invalid_fields(self, job, extra_validation_methods=None):
        """
        Return a list of all the invalid fields for the given job
        """
        invalid_fields = []

        plugin = job.get('plugin')
        if not plugin:
            logging.error("'plugin' field not found for job ({name})"
                          .format(name=job.get('name')))
            return invalid_fields

        name = job.get('name')
        if name:
            if any(character in string.whitespace for character in name):
                invalid_fields.append(('name',
                                       'Whitespace characters not allowed'))

        # Perform mandatory field validation
        for key in itertools.chain(self.MANDATORY_FIELDS,
                                   self.MANDATORY_PLUGIN_FIELDS.get(plugin, [])):
            field = job.get(key)
            if not field:
                logging.warning('{key!r} field not found for job {name!r}'
                                .format(key=key,
                                        name=job.get('name')))
                invalid_fields.append((key, ('{0!r} mandatory field not found'
                                             .format(key))))

        # Perform extra field validation
        # using the methods defined in the validation file
        if extra_validation_methods:
            for key, value in job.iteritems():
                if key in invalid_fields:
                    continue

                extra_validation_method = extra_validation_methods.get(key)
                if extra_validation_method:
                    try:
                        valid, reason = extra_validation_method(value, plugin)
                    except Exception as exception:
                        logging.error('Problem while executing {0!r} validation method:\n{1}'
                                      .format(key, exception))
                    else:
                        if not valid:
                            invalid_fields.append((key, reason))

        return invalid_fields


    def get_problems(self, model):
        """
        Look for errors
        """
        # Make sure there aren't any problems to be fixed
        problems = []

        # Include preferences and model problems
        problems.extend(model.preferences.problems)
        problems.extend(model.problems)

        for row in model.get_all_rows():
            description = row[model.DESCRIPTION]
            if description:
                extra_validation = getattr(model.preferences, 'extra_validation')
                invalid_fields = model.validator.get_invalid_fields(description,
                                                                    extra_validation)
                if invalid_fields:
                    for invalid_field, reason in invalid_fields:
                        problem = ('Invalid field',
                                   '{0!r} in field {1!r} for job {2!r}'
                                   .format(reason, invalid_field, description.get('name')))
                        problems.append(problem)

                        # Take care of the treeview coloring
                        model._set_problem_found(row.iter)

                name = description.get('name')
                if name:
                    if len(model.name_cache.get(name, [])) > 1:
                        problem = ('Duplicated name',
                                   '{0!r} is being used as name by multiple jobs'
                                   .format(name))
                        if not problem in problems:
                            problems.append(problem)
        return problems


    def display_problems(self, problems, standalone=False):
        """
        Display problems dialog
        """
        if standalone:
            message = ('Some <b>problems</b> have been found.\n'
                       'Please make sure to <b>fix</b> them before saving changes.')
            buttons = (gtk.STOCK_OK, gtk.RESPONSE_OK)
            default_response = gtk.RESPONSE_OK
        else:
            message = ('Some <b>problems</b> have been found.\n'
                       'Would you like to continue?.')
            buttons = (gtk.STOCK_YES, gtk.RESPONSE_YES,
                       gtk.STOCK_NO, gtk.RESPONSE_NO,
                       )
            default_response = gtk.RESPONSE_NO

        dialog = gtk.Dialog('Problems found', buttons=buttons)
        dialog.set_icon_name('checkbox-editor')
        dialog.set_default_response(default_response)
        dialog.set_property('border-width', 10)
        dialog.vbox.set_spacing(5)
        hbox = gtk.HBox(spacing=5)
        image = gtk.image_new_from_stock(gtk.STOCK_DIALOG_WARNING,
                                         gtk.ICON_SIZE_DIALOG)
        image.set_alignment(0.5, 0.0)
        label = gtk.Label()
        label.set_markup(message)
        hbox.pack_start(image, expand=False)
        hbox.pack_start(label, expand=False)
        dialog.vbox.pack_start(hbox, expand=False)

        expander = gtk.Expander('More information...')
        scrolled_window = gtk.ScrolledWindow()
        scrolled_window.set_policy(gtk.POLICY_AUTOMATIC,
                                   gtk.POLICY_AUTOMATIC)

        model = gtk.ListStore(str, str)
        treeview = gtk.TreeView(model)
        renderer = gtk.CellRendererText()
        column = gtk.TreeViewColumn('Problem Type', renderer, text=0)
        treeview.append_column(column)
        column = gtk.TreeViewColumn('Description', renderer, text=1)
        treeview.append_column(column)

        for problem in problems:
            model.append(problem)

        scrolled_window.add(treeview)
        expander.add(scrolled_window)

        dialog.vbox.pack_start(expander)
        dialog.show_all()

        if standalone:
            dialog.connect('response', lambda w, d: dialog.destroy())
        else:
            response = dialog.run()
            dialog.destroy()
            return response
