"""
version control support
"""
import pygtk
pygtk.require("2.0")
import gtk

import os, logging, threading, shutil
from cStringIO import StringIO

from contextlib import contextmanager

from bzrlib.workingtree import WorkingTree
from bzrlib.builtins import cmd_push as PushCommand
from bzrlib.branch import Branch
from bzrlib.errors import NotBranchError, BzrError
import bzrlib.plugin
bzrlib.plugin.load_plugins()

from .util import handle_exceptions, MessageDialogRunner


class VersionControl:
    """
    Base class for version control
    """
    @staticmethod
    def get(directory, exclude):
        """
        Get version control instance
        """
        try:
            vc = BazaarVC(directory, exclude)
        except NotBranchError:
            vc = NullVC(directory)

        return vc


class BazaarVC(VersionControl):
    """
    Bazaar version control
    """
    def __init__(self, directory, exclude):
        self.directory = directory
        if exclude is not None:
            self.exclude = exclude
        else:
            self.exclude = []
        self.working_tree = WorkingTree.open(directory)


    def __contains__(self, filename):
        path = os.path.relpath(filename, self.directory)
        return bool(self.working_tree.path2id(path))


    @contextmanager
    def _lock_read(self):
        """
        Acquire read lock, perform operations and release
        """
        self.working_tree.lock_read()
        yield
        self.working_tree.unlock()


    def has_changes(self):
        """
        Return tree if working tree has changes
        """
        with self._lock_read():
            w = self.working_tree
            filenames = (change[1][1]
                         for change in w.iter_changes(w.basis_tree()))
            changes_found = any(filenames)

        return changes_found


    def revno(self):
        """
        Return latest revision number
        """
        branch = Branch.open(self.directory)
        revno = branch.revno()
        logging.debug('{0!r} branch revision number is {1}'
                      .format(self.directory, revno))
        return revno


    def add(self, filename):
        """
        Add file under version control
        """
        logging.debug('Adding {0} under version control'.format(repr(filename)))
        self.working_tree.smart_add([filename])


    def remove(self, filename):
        """
        Remove file under version control
        """
        logging.debug('Removing {0} from version control'.format(repr(filename)))
        self.working_tree.remove([filename],
                                 keep_files=False, force=True)


    def rename(self, old_name, new_name):
        """
        Rename file under version control
        """
        logging.debug('Renaming from {0} to {1} under version control'
                      .format(repr(old_name), repr(new_name)))
        old_name = os.path.relpath(old_name, self.directory)
        new_name = os.path.relpath(new_name, self.directory)
        self.working_tree.rename_one(old_name, new_name)


    def commit(self):
        """
        Commit changes
        """
        # Default response
        self.committed = False

        # Gather branch changes
        with self._lock_read():
            w = self.working_tree
            bzr_changes = list(w.iter_changes(w.basis_tree()))
        changes = {'added': [],
                   'modified': [],
                   'renamed': [],
                   'removed': [],
                   }
        for bzr_change in bzr_changes:
            (_, (from_path, to_path), _,
             (from_versioned, to_versioned), _, _, _, _) = bzr_change

            # Filter out filenames that are being excluded from the commit
            if any(path in self.exclude for path in [from_path, to_path]):
                continue

            if not to_versioned:
                changes['removed'].append(from_path)
                continue

            if not from_versioned:
                changes['added'].append(to_path)
                continue

            if from_path != to_path:
                changes['renamed'].append('{0} => {1}'.format(from_path, to_path))
            else:
                changes['modified'].append(from_path)

        # Display bzr_changes
        if any(change for change in changes.values()):
            dialog = gtk.Dialog('Commit',
                                buttons = (gtk.STOCK_OK, gtk.RESPONSE_ACCEPT,
                                           gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT))
            dialog.set_property('border-width', 10)
            dialog.vbox.set_spacing(5)
            dialog.set_response_sensitive(gtk.RESPONSE_ACCEPT, False)

            hbox = gtk.HBox(spacing=5)
            image = gtk.image_new_from_stock(gtk.STOCK_DIALOG_QUESTION,
                                             gtk.ICON_SIZE_DIALOG)
            image.set_alignment(0.5, 0.0)
            label = gtk.Label('Please enter a description for the changes:')
            hbox.pack_start(image, expand=False)
            hbox.pack_start(label, expand=False)
            dialog.vbox.pack_start(hbox, expand=False)

            scrolled_window = gtk.ScrolledWindow()
            scrolled_window.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
            textview = gtk.TextView()
            textview.set_size_request(-1, 100)
            textbuffer = textview.get_buffer()
            textbuffer.connect('changed',
                               lambda tb: dialog.set_response_sensitive(gtk.RESPONSE_ACCEPT,
                                                                        bool(tb.get_property('text'))))

            scrolled_window.add(textview)
            dialog.vbox.pack_start(scrolled_window)

            model = gtk.TreeStore(str)
            treeview = gtk.TreeView(model)
            treeview.set_property('headers-visible', False)
            renderer = gtk.CellRendererText()
            column = gtk.TreeViewColumn('', renderer, text=0)
            treeview.append_column(column)

            root_row = model.append(None, ('Changes',))
            for change_type, changes_data in changes.items():
                if changes_data:
                    parent_row = model.append(root_row, (change_type.capitalize(),))

                    for change_data in changes_data:
                        model.append(parent_row, (change_data,))

            dialog.vbox.pack_start(treeview, expand=False)

            dialog.show_all()
            response = dialog.run()
            buffer = textview.get_buffer()
            commit_message = buffer.get_text(buffer.get_start_iter(),
                                             buffer.get_end_iter(),
                                             include_hidden_chars=False)
            dialog.destroy()

            if response == gtk.RESPONSE_ACCEPT:
                message = 'Committing changes:\n{0}'.format(commit_message)
                logging.debug(message)

                @handle_exceptions(BzrError,
                                   'Commit operation <b>failed</b>')
                def commit_target():
                    self.working_tree.commit(commit_message,
                                             allow_pointless=False,
                                             exclude=self.exclude)
                    self.committed = True
                thread = threading.Thread(target=commit_target)
                self._display_progress_dialog('Commit', message, thread)
        else:
            dialog = gtk.MessageDialog(flags=gtk.DIALOG_MODAL,
                                       type=gtk.MESSAGE_WARNING,
                                       buttons=gtk.BUTTONS_OK)
            dialog.set_title('Changes not to be committed')
            dialog.set_markup('This branch contains some changes\n'
                              'that are <b>excluded</b> from being committed\n\n'
                              'Please review version control tab\n'
                              'in preferences dialog to <b>modify</b>\n'
                              'excluded files list')

            content_area = dialog.get_content_area()
            expander = gtk.Expander('Excluded files...')
            content_area.pack_start(expander)

            model = gtk.ListStore(str)
            treeview = gtk.TreeView(model)
            treeview.set_property('headers-visible', False)
            renderer = gtk.CellRendererText()
            column = gtk.TreeViewColumn('', renderer, text=0)
            treeview.append_column(column)

            for excluded_file in self.exclude:
                model.append((excluded_file,))
            expander.add(treeview)

            dialog.show_all()
            dialog.run()
            dialog.destroy()

        return self.committed


    def push(self, branch_url):
        """
        Push changes
        """
        if not branch_url:
            MessageDialogRunner('Error',
                                'Push branch <b>not</b> defined '
                                'in preferences.').run()
            return False

        logging.debug('Pushing changes to: {0}'.format(repr(branch_url)))

        self.pushed = False
        @handle_exceptions(BzrError,
                           'Push operation <b>failed</b>')
        def push_target():
            cmd = PushCommand()
            cmd.outf = StringIO()
            cmd.run(location = branch_url,
                    directory = self.directory)
            self.pushed = True
        thread = threading.Thread(target=push_target)

        self._display_progress_dialog('Push',
                                      'Pushing changes to:\n<b>{0}</b>'.format(branch_url),
                                      thread)
        return self.pushed


    def pull(self, branch_url):
        """
        Pull changes
        """
        if not branch_url:
            MessageDialogRunner('Error',
                                'Pull branch <b>not</b> defined '
                                'in preferences.').run()
            return False

        logging.debug('Pulling changes from: {0!r}'.format(branch_url))

        self.pulled = False
        @handle_exceptions(BzrError,
                           'Pull operation <b>failed</b>')
        def pull_target():
            branch = Branch.open(branch_url)
            self.pull_result = self.working_tree.pull(branch)
            self.pulled = True
        thread = threading.Thread(target=pull_target)

        self._display_progress_dialog('Pull',
                                      'Pulling changes from:\n<b>{0}</b>'.format(branch_url),
                                      thread)

        if self.pulled and not self._check_conflicts():
            return True

        return False


    def _check_conflicts(self):
        """
        Check for conflicts in the branch
        """
        conflicts = self.working_tree.conflicts()
        if conflicts.is_empty():
            return False

        conflicts_str = '\n'.join([conflict.path for conflict in conflicts])
        logging.warning('Conflicts were found:\n{0}'.format(conflicts_str))

        dialog = gtk.Dialog('Conflicts',
                            flags = gtk.DIALOG_MODAL,
                            buttons = (gtk.STOCK_OK, gtk.RESPONSE_OK))
        dialog.set_property('border-width', 10)
        hbox = gtk.HBox(spacing=5)
        image = gtk.image_new_from_stock(gtk.STOCK_DIALOG_WARNING,
                                         gtk.ICON_SIZE_DIALOG)
        image.set_alignment(0.5, 0.0)
        label = gtk.Label('Conflicts were found')
        hbox.pack_start(image, expand=False)
        hbox.pack_start(label, expand=False)
        dialog.vbox.pack_start(hbox, expand=False)

        expander = gtk.Expander('Files affected')
        scrolled_window = gtk.ScrolledWindow()
        scrolled_window.set_policy(gtk.POLICY_AUTOMATIC,
                                   gtk.POLICY_AUTOMATIC)
        textview = gtk.TextView()
        textview.set_editable(False)
        buffer = textview.get_buffer()
        buffer.set_text(conflicts_str)
        scrolled_window.add(textview)
        expander.add(scrolled_window)

        dialog.vbox.pack_start(expander)
        dialog.connect('response', lambda d, r: d.destroy())
        dialog.show_all()
        return True


    def _display_progress_dialog(self, title, message, thread):
        """
        Display a progress dialog while the version control operation is bein performed
        """
        dialog = gtk.Dialog(title,
                            flags = gtk.DIALOG_MODAL)
        dialog.set_icon_name('checkbox-editor')
        dialog.set_property('border-width', 10)
        dialog.set_has_separator(False)
        dialog.vbox.set_spacing(5)

        hbox = gtk.HBox(spacing=5)
        image = gtk.image_new_from_stock(gtk.STOCK_DIALOG_INFO,
                                         gtk.ICON_SIZE_DIALOG)
        image.set_alignment(0.5, 0.0)
        hbox.pack_start(image)
        label = gtk.Label()
        label.set_markup(message)
        hbox.pack_start(label)
        dialog.vbox.pack_start(hbox)

        progress_bar = gtk.ProgressBar()
        dialog.vbox.pack_start(progress_bar)

        # Prevent dialog from being closed and resized
        dialog.connect('delete-event', lambda w, e: True)
        dialog.set_deletable(False)
        dialog.set_resizable(False)
        dialog.show_all()

        def refresh():
            """
            Process pending events
            """
            while gtk.events_pending():
                gtk.main_iteration()

        refresh()

        thread.start()
        while True:
            thread.join(0.2)
            if not thread.is_alive():
                break

            progress_bar.pulse()
            refresh()

        dialog.destroy()


    def clone(self, destination_directory):
        """
        Clone working directory to destination
        """
        logging.debug('Cloning bzr repository from {0!r} to {1!r}'
                      .format(self.directory, destination_directory))
        self.working_tree.bzrdir.sprout(destination_directory)


class NullVC(VersionControl):
    """
    No version control at all, just files
    """
    def __init__(self, directory):
        self.directory = directory


    def __contains__(self, filename):
        return os.path.isfile(filename)


    def has_changes(self):
        return False


    def add(self, filename):
        """
        Add file to the filesystem
        """
        # Do nothing since file is written
        # before calling the version control manager
        pass


    def remove(self, filename):
        """
        Remove file from the filesystem
        """
        os.remove(filename)


    def rename(self, old_name, new_name):
        """
        Rename file in the file system
        """
        os.rename(old_name, new_name)


    def clone(self, destination_directory):
        """
        Copy all files to a remote directory
        """
        # Directory must not exist for copytree to work
        # but FileChooserDialog creates it, so here it's
        # removed before trying to copy the contents
        # of the original directory
        os.rmdir(destination_directory)

        # Copy all files in the old shared directory to the new one
        source_directory = self.directory
        logging.debug('Copying files from {0!r} to {1!r}'
                      .format(source_directory, destination_directory))
        shutil.copytree(source_directory, destination_directory, symlinks=True)
