"""
whitelist with job ordering implementation
"""
import pygtk
pygtk.require("2.0")
import gtk

import re, logging
from datetime import datetime

from .data import Reporter
from .util import handle_exceptions, MessageDialogRunner
from . import build


class Whitelist(object):
    """
    Whitelist object is an ordered list of jobs
    """
    BUTTON_NAMES = ('whitelist_top_button', 'whitelist_up_button',
                    'whitelist_down_button', 'whitelist_bottom_button')


    def __init__(self, main_model):
        self.main_model = main_model
        self.builder = main_model.builder
        self.treeview = self.builder.get_object('whitelist_treeview')
        self.model = self.treeview.get_model()

        selection = self.treeview.get_selection()
        selection.set_mode(gtk.SELECTION_SINGLE)
        selection.connect('changed', self.selection_changed_cb)

        self.filename = None
        self.references = {}


    def clear(self):
        """
        Initialize internal data
        """
        self.model.clear()
        self.filename = None
        self.references = {}


    def append(self, name):
        """
        Append a job name to the whitelist
        """
        # If name is already in the whitelist
        # probably it's a duplicated name entry
        # so it safest just to keep the name
        # in its current location to avoid breaking
        # ordering dependencies
        if name in self.references:
            return

        iterator = self.model.append((name,))
        path = self.model.get_path(iterator)
        reference = gtk.TreeRowReference(self.model, path)
        self.references[name] = reference

        selection = self.treeview.get_selection()
        self._update_buttons_sensitivity(selection)


    def remove(self, name):
        """
        Remove a job name from the whitelist
        """
        # Make sure that job name isn't selected multiple
        # times before removing
        # (this might happen if name is duplicated)
        references = self.main_model.name_cache[name]
        rows = [self.main_model[reference.get_path()]
                for reference in references]
        rows = [row for row in rows if row[self.main_model.SELECTED]]
        if len(rows) > 1:
            return

        if name in self.references:
            reference = self.references.pop(name)
            path = reference.get_path()
            iterator = self.model.get_iter(path)
            self.model.remove(iterator)

        selection = self.treeview.get_selection()
        self._update_buttons_sensitivity(selection)


    def selection_changed_cb(self, selection):
        """
        Update buttons sensitivity
        """
        self._update_buttons_sensitivity(selection)


    def top_button_clicked_cb(self, button):
        """
        Move job to the top of the whitelist
        """
        selection = self.treeview.get_selection()
        model, selected_iterator = selection.get_selected()

        # Gather all dependencies (including ancestors)
        selected_row = model[selected_iterator]
        selected_name = selected_row[0]
        dependencies = self._get_dependencies(selected_name)

        previous_rows = []
        previous_row = model[model.get_iter_first()]
        while previous_row.path != selected_row.path:
            name = previous_row[0]
            if name not in dependencies:
                previous_rows.append(previous_row)
            previous_row = previous_row.next

        if not previous_rows:
            message = ('{0} <b>cannot</b> be moved up without '
                       'breaking any dependency'.format(selected_name))
            MessageDialogRunner('Whitelist ordering', message,
                                gtk.MESSAGE_WARNING).run()
            return

        for row in reversed(previous_rows):
            model.move_after(row.iter, selected_iterator)

        self._update_buttons_sensitivity(selection)
        self._whitelist_changed()


    def up_button_clicked_cb(self, button):
        """
        Move job up in the whitelist
        """
        selection = self.treeview.get_selection()
        model, selected_iterator = selection.get_selected()

        # Gather all dependencies (including ancestors)
        selected_row = model[selected_iterator]
        selected_name = selected_row[0]
        dependencies = self._get_dependencies(selected_name)

        # Move previous row after selected row
        # (look for other previous rows to honor dependencies)
        previous_path = model.get_path(selected_iterator)[0]-1
        previous_iterator = model.get_iter(previous_path)
        previous_row = model[previous_iterator]
        previous_name = previous_row[0]
        while previous_name in dependencies:
            previous_path -= 1
            if previous_path < 0:
                message = ('{0} <b>cannot</b> be moved up without '
                           'breaking any dependency'.format(selected_name))
                MessageDialogRunner('Whitelist ordering', message,
                                    gtk.MESSAGE_WARNING).run()
                return
            previous_iterator = model.get_iter(previous_path)
            previous_row = model[previous_iterator]
            previous_name = previous_row[0]

        model.move_after(previous_iterator, selected_iterator)

        self._update_buttons_sensitivity(selection)
        self._whitelist_changed()


    def down_button_clicked_cb(self, button):
        """
        Move job down in the whitelist
        """
        selection = self.treeview.get_selection()
        model, selected_iterator = selection.get_selected()
        selected_row = model[selected_iterator]
        selected_name = selected_row[0]

        next_row = model[selected_iterator].next
        name = next_row[0]
        dependencies = self._get_dependencies(name)
        while selected_name in dependencies:
            next_row = next_row.next
            if not next_row:
                message = ('{0} <b>cannot</b> be moved down without '
                           'breaking any dependency'.format(selected_name))
                MessageDialogRunner('Whitelist ordering', message,
                                    gtk.MESSAGE_WARNING).run()
                return

            name = next_row[0]
            dependencies = self._get_dependencies(name)

        next_iterator = next_row.iter
        model.move_before(next_iterator, selected_iterator)

        self._update_buttons_sensitivity(selection)
        self._whitelist_changed()


    def bottom_button_clicked_cb(self, button):
        """
        Move job to the bottom of the whitelist
        """
        selection = self.treeview.get_selection()
        model, selected_iterator = selection.get_selected()
        selected_row = model[selected_iterator]
        selected_name = selected_row[0]

        next_rows = []
        next_row = selected_row.next
        while next_row:
            name = next_row[0]
            if selected_name not in self._get_dependencies(name):
                next_rows.append(next_row)
            next_row = next_row.next

        if not next_rows:
            message = ('{0} <b>cannot</b> be moved down without '
                       'breaking any dependency'.format(selected_name))
            MessageDialogRunner('Whitelist ordering', message,
                                gtk.MESSAGE_WARNING).run()
            return

        for row in next_rows:
            model.move_before(row.iter, selected_iterator)

        self._update_buttons_sensitivity(selection)
        self._whitelist_changed()


    def _update_buttons_sensitivity(self, selection):
        """
        Set buttons sensitivity according to current selection
        """
        model, iterator = selection.get_selected()

        if iterator:
            row = model[iterator]

            sensitivity = bool(row.path[0])
            for widget_name in ('whitelist_top_button',
                                'whitelist_up_button'):
                widget = self.builder.get_object(widget_name)
                widget.set_sensitive(sensitivity)

            sensitivity = bool(row.next)
            for widget_name in ('whitelist_down_button',
                                'whitelist_bottom_button'):
                widget = self.builder.get_object(widget_name)
                widget.set_sensitive(sensitivity)
        else:
            for widget_name in self.BUTTON_NAMES:
                widget = self.builder.get_object(widget_name)
                widget.set_sensitive(False)


    def _whitelist_changed(self):
        """
        Update save buttons sensitivity
        """
        widget_names = ['save_whitelist_as_menuitem']

        if self.filename:
            widget_names.append('save_whitelist_menuitem')

        for widget_name in widget_names:
            widget = self.builder.get_object(widget_name)
            widget.set_sensitive(True)


    def _get_dependencies(self, name):
        """
        Get all dependencies including ancestors
        """
        dependencies = set()
        references = self.main_model.name_cache[name]

        for reference in references:
            row = self.main_model[reference.get_path()]
            if not row[self.main_model.SELECTED]:
                continue

            description = row[self.main_model.DESCRIPTION]
            depends = description.get('depends')
            if depends:
                dependencies.add(depends)
                dependencies.update(self._get_dependencies(depends))

            parent_row = row.parent
            while parent_row:
                description = parent_row[self.main_model.DESCRIPTION]
                if description:
                    name = description.get('name')
                    if name:
                        dependencies.add(name)
                parent_row = parent_row.parent

        return dependencies


    @handle_exceptions(re.error,
                       'Load Whitelist operation <b>failed</b>')
    def load(self, filename):
        """
        Load a whitelist file and apply patterns to select jobs
        """
        logging.debug('Applying whitelist: {0}'.format(filename))

        # Clear before loading
        self.clear()

        # Gather all patterns
        # If an illegal expresion is in the file,
        # an exception will be raised before any change in model
        patterns = []
        with open(filename) as f:
            for line in f:
                if line.startswith('#'):
                    continue
                pattern_str = '^{0}$'.format(line.rstrip('\n'))
                pattern = re.compile(pattern_str)
                patterns.append(pattern)

        # Clear all previous selections
        for row in self.main_model.get_all_rows():
            row[self.main_model.SELECTED] = False

        # Apply patterns loaded from file
        unused_patterns = []
        for pattern in patterns:
            pattern_matched = False

            for row in self.main_model.get_all_rows():
                description = row[self.main_model.DESCRIPTION]
                if (description and 'name' in description
                    and pattern.match(description['name'])):
                    logging.debug('{0!r} selected because of pattern {1!r}'
                                  .format(description['name'], pattern.pattern))
                    self.append(description['name'])
                    row[self.main_model.SELECTED] = True
                    pattern_matched = True

            if not pattern_matched:
                unused_patterns.append(pattern.pattern[1:-1])

        if unused_patterns:
            message = ['The following patterns did <b>not</b> match any of the jobs '
                       'and will be <b>ignored</b>:',
                       '']
            for pattern in unused_patterns:
                message.append('- <b>{0}</b>'.format(pattern))
            message = '\n'.join(message)
            MessageDialogRunner('Unused patterns', message,
                                gtk.MESSAGE_WARNING).run()

        # Final trick is select root nodes (filenames) if they have selected children
        # This is just because is confusing not to select a file
        # when children are selected, but this isn't something
        # that's part of the whitelist file that checkbox uses
        for row in self.main_model:
            if any(child_row[self.main_model.SELECTED]
                   for child_row in row.iterchildren()):
                row[self.main_model.SELECTED] = True

        # Remember filename
        self.filename = filename
        for widget_name in ('save_whitelist_menuitem',
                            'save_whitelist_as_menuitem'):
            widget = self.builder.get_object(widget_name)
            widget.set_sensitive(False)

        anything_selected = any(self.main_model.get_selected_rows())
        for widget_name in ('execute_menuitem', 'execute_toolbutton',
                            'export_whitelist_menuitem'):
            widget = self.builder.get_object(widget_name)
            widget.set_sensitive(anything_selected)

        # Check that for every node in the tree all its ancestors are selected
        orphan_rows = []
        for row in self.main_model.get_selected_rows():
            parent_row = row.parent
            while parent_row:
                if not parent_row[self.main_model.SELECTED]:
                    orphan_rows.append(row)
                    break
                parent_row = parent_row.parent

        if orphan_rows:
            message = ['The following jobs have been <b>whitelisted</b>, '
                       'but some of their ancestors have <b>not</b>:',
                       '']
            for row in orphan_rows:
                message.append('- <b>{0}</b>'.format(row[self.main_model.DISPLAY_NAME]))
            message.extend(['',
                            'The problem will be fixed by selecting all ancestors',
                            'that were missing.'])
            message = '\n'.join(message)
            MessageDialogRunner('Orphan jobs', message,
                                gtk.MESSAGE_WARNING).run()

            for row in orphan_rows:
                self.main_model.set_selected(row, True)


    @handle_exceptions(IOError,
                       'Save Whitelist operation <b>failed</b>')
    def save(self, filename = None):
        """
        Save whitelist to file
        """
        if filename is None:
            filename = self.filename

        selected_patterns = []
        for name, in self.model:
            references = self.main_model.name_cache[name]
            rows = (self.main_model[reference.get_path()]
                    for reference in references)
            descriptions = (row[self.main_model.DESCRIPTION]
                            for row in rows)
            plugins = (description.get('plugin')
                       for description in descriptions)

            # Remote plugins are expected to generate job names
            # that start with the same prefix as the 'remote'
            # job used to generate them
            if any(plugin.startswith('remote')
                   for plugin in plugins):
                pattern = '{0}.*'
            else:
                pattern = '{0}'

            selected_patterns.append(pattern.format(re.escape(name)))

        logging.info('Saving whitelist to {0!r}...'.format(filename))
        logging.debug('Selected patterns:\n{0}'
                      .format('\n'.join(selected_patterns)))

        # Information that will added as a header to the whitelist
        header = ['# Checkbox whitelist file created by checkbox-editor {0}'
                  .format(build.version),
                  '# Date: {0}'.format(datetime.now().strftime('%Y-%m-%d %H:%M:%S')),
                  ]

        vc = self.main_model.vc
        if hasattr(vc, 'revno'):
            header.append('# Branch: {0} (rev. {1})'.format(vc.directory,
                                                            vc.revno()))
        else:
            header.append('# Branch: {0}'.format(vc.directory))

        if hasattr(self.main_model.vc, 'pull'):
            pull_branch = self.main_model.preferences.version_control.get('pull_branch', None)
            if pull_branch:
                header.append('# Pull branch: {0}'.format(pull_branch))

        with open(filename, 'w') as f:
            header.append('')
            f.write('\n'.join(header))
            lines = ('{0}\n'.format(pattern)
                     for pattern in selected_patterns)
            f.writelines(lines)

        # Remember filename
        # (in case file name was modified using 'save as' menuitem)
        self.filename = filename

        logging.info('Whitelist successfully saved')


    @handle_exceptions(IOError,
                       'Export Whitelist operation <b>failed</b>')
    def export(self, report_type, destination):
        """
        Export whitelist to a file
        """
        selected_description_tree = self.main_model.get_selected_description_tree()

        reporter = Reporter.get(report_type, destination, selected_description_tree)
        reporter.write_report()
